#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#




"""Sends email on behalf of the application.

This module provides functions for application developers to provide email
services for their applications. The module also provides a few utility methods.
"""










import binascii
import cgi
import codecs
from collections.abc import MutableMapping
import email
from email import parser
import email.header
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
import functools
import logging
import typing

from google.appengine.api import api_base_pb2
from google.appengine.api import apiproxy_stub_map
from google.appengine.api import mail_service_pb2
from google.appengine.api import users
from google.appengine.api.mail_errors import *
from google.appengine.runtime import apiproxy_errors
import six
from six.moves import map










ERROR_MAP = {
    mail_service_pb2.MailServiceError.BAD_REQUEST:
      BadRequestError,
    mail_service_pb2.MailServiceError.UNAUTHORIZED_SENDER:
      InvalidSenderError,
    mail_service_pb2.MailServiceError.INVALID_ATTACHMENT_TYPE:
      InvalidAttachmentTypeError,
    mail_service_pb2.MailServiceError.INVALID_HEADER_NAME:
      InvalidHeaderNameError,
}







EXTENSION_MIME_MAP = {
    'aif': 'audio/x-aiff',
    'aifc': 'audio/x-aiff',
    'aiff': 'audio/x-aiff',
    'asc': 'text/plain',
    'au': 'audio/basic',
    'avi': 'video/x-msvideo',
    'bmp': 'image/x-ms-bmp',
    'c': 'text/plain',
    'css': 'text/css',
    'csv': 'text/csv',
    'doc': 'application/msword',
    'docx': 'application/msword',
    'diff': 'text/plain',
    'flac': 'audio/flac',
    'gif': 'image/gif',
    'gzip': 'application/x-gzip',
    'htm': 'text/html',
    'html': 'text/html',
    'ics': 'text/calendar',
    'jpe': 'image/jpeg',
    'jpeg': 'image/jpeg',
    'jpg': 'image/jpeg',
    'kml': 'application/vnd.google-earth.kml+xml',
    'kmz': 'application/vnd.google-earth.kmz',
    'log': 'text/plain',
    'm4a': 'audio/mp4',
    'mid': 'audio/mid',
    'mov': 'video/quicktime',
    'mp3': 'audio/mpeg',
    'mp4': 'video/mp4',
    'mpe': 'video/mpeg',
    'mpeg': 'video/mpeg',
    'mpg': 'video/mpeg',
    'odp': 'application/vnd.oasis.opendocument.presentation',
    'ods': 'application/vnd.oasis.opendocument.spreadsheet',
    'odt': 'application/vnd.oasis.opendocument.text',
    'oga': 'audio/ogg',
    'ogg': 'audio/ogg',
    'ogv': 'video/ogg',
    'patch': 'text/plain',
    'pdf': 'application/pdf',
    'png': 'image/png',
    'pot': 'text/plain',
    'pps': 'application/vnd.ms-powerpoint',
    'ppt': 'application/vnd.ms-powerpoint',
    'pptx': 'application/vnd.ms-powerpoint',
    'qt': 'video/quicktime',
    'rmi': 'audio/mid',
    'rss': 'text/rss+xml',
    'snd': 'audio/basic',
    'sxc': 'application/vnd.sun.xml.calc',
    'sxw': 'application/vnd.sun.xml.writer',
    'text': 'text/plain',
    'tif': 'image/tiff',
    'tiff': 'image/tiff',
    'txt': 'text/plain',
    'vcf': 'text/directory',
    'wav': 'audio/x-wav',
    'wbmp': 'image/vnd.wap.wbmp',
    'webm': 'video/webm',
    'webp': 'image/webp',
    'xls': 'application/vnd.ms-excel',
    'xlsx': 'application/vnd.ms-excel',
    'zip': 'application/zip',
}





EXTENSION_BLACKLIST = [
    'ade',
    'adp',
    'bat',
    'chm',
    'cmd',
    'com',
    'cpl',
    'exe',
    'hta',
    'ins',
    'isp',
    'jse',
    'lib',
    'mde',
    'msc',
    'msp',
    'mst',
    'pif',
    'scr',
    'sct',
    'shb',
    'sys',
    'vb',
    'vbe',
    'vbs',
    'vxd',
    'wsc',
    'wsf',
    'wsh',
]


HEADER_WHITELIST = frozenset([
    'Auto-Submitted',
    'In-Reply-To',
    'List-Id',
    'List-Unsubscribe',
    'On-Behalf-Of',
    'References',
    'Resent-Date',
    'Resent-From',
    'Resent-To',
])

INCOMING_MAIL_URL_PATTERN = '/_ah/mail/.+'
BOUNCE_NOTIFICATION_URL_PATH = '/_ah/bounce'


def invalid_email_reason(email_address, field):
  """Determines the reason why an email is invalid.

  Args:
    email_address: Email address to check.
    field: Field that is invalid.

  Returns:
    A string that indicates the reason why an email is invalid; otherwise
    returns `None`.
  """
  if email_address is None:
    return 'None email address for %s.' % field

  if isinstance(email_address, users.User):
    email_address = email_address.email()
  if not isinstance(email_address, six.string_types):
    return 'Invalid email address type for %s.' % field
  stripped_address = email_address.strip()
  if not stripped_address:
    return 'Empty email address for %s.' % field
  return None


InvalidEmailReason = invalid_email_reason


def is_email_valid(email_address):
  """Determines whether an email address is invalid.

  Args:
    email_address: Email address to check.

  Returns:
    `True` if the specified email address is valid; otherwise returns `False`.
  """
  return invalid_email_reason(email_address, '') is None


IsEmailValid = is_email_valid


def check_email_valid(email_address, field):
  """Verifies whether an email is valid.

  Args:
    email_address: Email address to check.
    field: Field to check.

  Raises:
    InvalidEmailError: If `email_address` is invalid.
  """
  reason = invalid_email_reason(email_address, field)
  if reason is not None:
    raise InvalidEmailError(reason)


CheckEmailValid = check_email_valid


def is_ascii(string):
  """Returns whether a string is in ASCII."""
  return all(ord(c) < 128 for c in string)


def invalid_headers_reason(headers):
  """Determines the reason why a header is invalid.

  Args:
    headers: Header value to check.

  Returns:
    A string that indicates the reason that the headers are invalid if the
    reason can be determined; otherwise returns `None`.
  """
  if headers is None:
    return 'Headers dictionary was None.'
  if not isinstance(headers, dict):
    return 'Invalid type for headers. Should be a dictionary.'
  for k, v in six.iteritems(headers):
    if not isinstance(k, six.string_types):
      return 'Header names should be strings.'
    if not isinstance(v, six.string_types):
      return 'Header values should be strings.'
    if not is_ascii(k):
      return 'Header name should be an ASCII string.'

    if k.strip() not in HEADER_WHITELIST:
      return 'Header "%s" is not allowed.' % k.strip()


def check_headers_valid(headers):
  """Checks that `headers` is a valid dictionary for a header.

  Args:
    headers: The value to check for the headers.

  Raises:
    InvalidEmailError: If `headers` is invalid.
  """
  reason = invalid_headers_reason(headers)
  if reason is not None:
    raise InvalidEmailError(reason)


def _email_sequence(emails):
  """Forces an email to be a sequenceable type.

  Iterable values are returned as-is. This function really just wraps the case
  where there is a single email string.

  Args:
    emails: At least one email address to coerce to sequence.

  Returns:
    A single tuple that contains the email address if only one email string is
    provided; otherwise returns email addresses as-is.
  """
  if isinstance(emails, six.string_types):
    return [email_address.strip() for email_address in emails.split(',')]
  return emails


def _attachment_sequence(attachments):
  """Forces attachments to be a sequenceable type.

  Iterable values are returned as-is. This function really just wraps the case
  where there is a single attachment.

  Args:
    attachments: Attachments (or attachment) to coerce to sequence.

  Yields:
    A single tuple that contains the attachment if only one attachment is
    provided; otherwise returns attachments as-is.
  """
  if len(attachments) == 2 and isinstance(attachments[0], six.string_types):
    attachments = attachments,
  for attachment in attachments:
    if isinstance(attachment, Attachment):
      yield attachment
    else:
      yield Attachment(*attachment)


def _parse_mime_message(
    mime_message: typing.Union[typing.Text, bytes, typing.TextIO,
                               email.message.Message]
) -> email.message.Message:
  """Helper function that converts `mime_message` into `email.message.Message`.

  Args:
    mime_message: MIME message, string, or file that contains the MIME
    message.

  Returns:
    An instance of `email.message.Message`. This method will return
    the `mime_message` if the instance already exists.
  """
  if isinstance(mime_message, email.message.Message):
    return mime_message
  elif isinstance(mime_message, six.string_types):

    return email.message_from_string(mime_message)
  elif isinstance(mime_message, bytes):
    return email.message_from_bytes(mime_message)
  else:

    return email.message_from_file(mime_message)


def send_mail(sender,
              to,
              subject,
              body,
              make_sync_call=apiproxy_stub_map.MakeSyncCall,
              **kw):
  """Sends mail on behalf of the application.

  Args:
    sender: Sender email address as it appears in the 'From' email line.
    to: List of one or more 'To' addresses.
    subject: Message subject string.
    body: Plain-text body.
    make_sync_call: Function used to make a sync call to an API proxy.
    **kw: Keyword arguments that are compatible with the `EmailMessage`
        keyword based constructor.

  Raises:
    InvalidEmailError: If an invalid email address was specified.
  """
  kw['sender'] = sender
  kw['to'] = to
  kw['subject'] = subject
  kw['body'] = body
  message = EmailMessage(**kw)
  message.send(make_sync_call=make_sync_call)


SendMail = send_mail


def send_mail_to_admins(sender,
                        subject,
                        body,
                        make_sync_call=apiproxy_stub_map.MakeSyncCall,
                        **kw):
  """Sends email to administrators on behalf of the application.

  Args:
    sender: Sender email address as it appears in the 'From' email line.
    subject: Message subject string.
    body: Plain-text body.
    make_sync_call: Function used to make a sync call to an API proxy.
    **kw: Keyword arguments that are compatible with the `EmailMessage` keyword
        based constructor.

  Raises:
    InvalidEmailError: If an invalid email address was specified.
  """
  kw['sender'] = sender
  kw['subject'] = subject
  kw['body'] = body
  message = AdminEmailMessage(**kw)
  message.send(make_sync_call=make_sync_call)


SendMailToAdmins = send_mail_to_admins


def _GetMimeType(file_name):
  """Determines the `MINE` type from the file name.

  This function parses the file name and determines the MIME type based on
  an extension map.

  This method is not part of the public API and should not be used by
  applications.

  Args:
    file_name: File for which you are attempting to determine the extension.

  Returns:
    The MIME type that is associated with the file extension.

  Raises:
    InvalidAttachmentTypeError: If the file type is invalid.
  """
  extension_index = file_name.rfind('.')
  if extension_index == -1:
    extension = ''
  else:
    extension = file_name[extension_index + 1:].lower()
  if extension in EXTENSION_BLACKLIST:
    raise InvalidAttachmentTypeError(
        'Extension %s is not supported.' % extension)
  mime_type = EXTENSION_MIME_MAP.get(extension, None)
  if mime_type is None:
    mime_type = 'application/octet-stream'
  return mime_type


def _GuessCharset(text):
  """Guesses the character set of a piece of text.

  Args:
    text: A string that is either a US-ASCII string or a Unicode string that was
        encoded in UTF-8.

  Returns:
    The character set that is needed by the string, either US-ASCII or UTF-8.
  """
  if isinstance(text, six.text_type):
    try:
      text.encode('us-ascii')
      return 'us-ascii'
    except UnicodeEncodeError:
      return 'utf-8'

  else:
    try:
      text.decode('us-ascii')
      return 'us-ascii'
    except UnicodeDecodeError:
      return 'utf-8'


def _I18nHeader(text):
  """Creates a properly encoded header, even with Unicode content.

  Args:
    text: A string that is either a US-ASCII string or a Unicode string that was
        encoded in UTF-8.
  Returns:
    The properly encoded `email.header.Header`.
  """
  charset = _GuessCharset(text)
  return email.header.Header(text, charset, maxlinelen=1e3000)


def mail_message_to_mime_message(protocol_message):
  """Generates a `MIMEMultipart` message from a `MailMessage` protocol buffer.

  This function generates a complete `MIMEMultipart` email object from a
  `MailMessage` protocol buffer. The body fields are sent as individual
  alternatives if they are both present; otherwise, only one body part is sent.

  Multiple entry email fields, such as 'To', 'Cc', and 'Bcc' are converted
  to a list of comma-separated email addresses.

  Args:
    protocol_message: Message protocol buffer to convert to a `MIMEMultipart`
        message.

  Returns:
    A `MIMEMultipart` message that represents the provided `MailMessage`.

  Raises:
    InvalidAttachmentTypeError: If the file type of the attachment is invalid.
  """
  parts = []
  if protocol_message.HasField('TextBody'):
    parts.append(
        MIMEText(
            protocol_message.TextBody,
            _charset=_GuessCharset(protocol_message.TextBody)))
  if protocol_message.HasField('AmpHtmlBody'):
    parts.append(
        MIMEText(
            protocol_message.AmpHtmlBody,
            _subtype='x-amp-html',
            _charset=_GuessCharset(protocol_message.AmpHtmlBody)))
  if protocol_message.HasField('HtmlBody'):
    parts.append(
        MIMEText(
            protocol_message.HtmlBody,
            _subtype='html',
            _charset=_GuessCharset(protocol_message.HtmlBody)))

  if len(parts) == 1:

    payload = parts
  else:

    payload = [MIMEMultipart('alternative', _subparts=parts)]

  result = MIMEMultipart(_subparts=payload)

  for attachment in protocol_message.Attachment:
    file_name = attachment.FileName
    mime_type = _GetMimeType(file_name)
    maintype, subtype = mime_type.split('/')
    mime_attachment = MIMEBase(maintype, subtype)
    mime_attachment.add_header(
        'Content-Disposition', 'attachment', filename=attachment.FileName)
    mime_attachment.set_payload(attachment.Data)
    if attachment.HasField('ContentID'):
      mime_attachment['content-id'] = attachment.ContentID
    result.attach(mime_attachment)


  if protocol_message.To:
    result['To'] = _I18nHeader(', '.join(protocol_message.To))
  if protocol_message.Cc:
    result['Cc'] = _I18nHeader(', '.join(protocol_message.Cc))
  if protocol_message.Bcc:
    result['Bcc'] = _I18nHeader(', '.join(protocol_message.Bcc))

  result['From'] = _I18nHeader(protocol_message.Sender)
  result['Reply-To'] = _I18nHeader(protocol_message.ReplyTo)
  result['Subject'] = _I18nHeader(protocol_message.Subject)

  for header in protocol_message.Header:
    result[header.name] = _I18nHeader(header.value)

  return result


MailMessageToMIMEMessage = mail_message_to_mime_message


def _to_str(value):
  """Helper function to ensure that Unicode values are converted to UTF-8.

  Args:
    value: String or Unicode to convert to UTF-8.

  Returns:
    The UTF-8 encoded string of `value`; `value` remains otherwise unchanged.
  """
  if isinstance(value, six.text_type):
    return value.encode('utf-8')
  return value


def _decode_and_join_header(header, separator=u' '):
  """Helper function to decode RFC2047 encoded headers.

  Args:
    header: `RFC2047`_ encoded string (or just a standard string) to convert to
        Unicode.
    separator: The separator to use when joining separately encoded pieces of
        the header.

  Returns:
    The Unicode version of the decoded header; returns `None` or `''` if the
    header is not set.

  .. _RFC2047:
     https://www.ietf.org/rfc/rfc2047.txt
  """
  if not header:

    return header

  def header_gen(header_string):
    """Generate a sequence of header from header string.

    A generator that consume header string and return a sequence of header in
    decoded format (PY2: str; PY3: unicode).

    Args:
      header_string: A string that contains multiline of headers.

    Yields:
      string
    """


    for s, charset in email.header.decode_header(header_string):
      if not charset:
        charset = 'us-ascii'



      if isinstance(s, bytes):
        yield codecs.decode(s, charset, 'replace')
      else:
        yield s

  return separator.join(header_gen(header))


def _decode_address_list_field(address_list):
  """Helper function to decode (potentially RFC2047 encoded) address lists.

  Args:
    address_list: A single string header, or list of string headers.

  Returns:
    The Unicode version of a decoded header or a list of string headers.
  """
  if not address_list:
    return None

  if len(address_list) == 1:
    return _decode_and_join_header(address_list[0])
  else:
    return list(map(_decode_and_join_header, address_list))




def _wrapping(wrapped):
  """A decorator that decorates a decorator's wrapper.

  This decorator makes it easier to debug code that is heavily decorated.

  Args:
    wrapped: The decorated function that you are trying to debug.

  Returns:
    A function with `__name__`, `__doc__`, and `___dict__` remapped to the
    respective versions of the wrapped function to make debugging easier.
  """




  def wrapping_wrapper(wrapper):
    try:
      wrapper.__wrapped__ = wrapped
      wrapper.__name__ = wrapped.__name__
      wrapper.__doc__ = wrapped.__doc__
      wrapper.__dict__.update(wrapped.__dict__)
    except Exception:
      pass
    return wrapper
  return wrapping_wrapper








wrapping = _wrapping if six.PY2 else functools.wraps



def _positional(max_pos_args):
  """A decorator to declare that only the first N arguments can be positional.

  Note:
      For methods, N includes 'self'.

  Args:
    max_pos_args: The number of arguments that can be positional.

  Returns:
    The wrapped object, but verifies that the wrapped object does not have more
    than `max_pos_args`. If the object does contain more than `max_pos_args`
    arguments, a `TypeError` is raised.
  """
  def positional_decorator(wrapped):
    @functools.wraps(wrapped)
    def positional_wrapper(*args, **kwds):
      if len(args) > max_pos_args:
        plural_s = ''
        if max_pos_args != 1:
          plural_s = 's'
        raise TypeError(
            '%s() takes at most %d positional argument%s (%d given)' %
            (wrapped.__name__, max_pos_args, plural_s, len(args)))
      return wrapped(*args, **kwds)
    return positional_wrapper
  return positional_decorator


class Attachment(object):
  """Attachment object.

  An Attachment object is largely interchangeable with a `(filename, payload)`
  tuple.

  Note:
      The behavior is a bit asymmetric with respect to unpacking and equality
      comparison. An Attachment object without a content ID will be equivalent
      to a `(filename, payload)` tuple. An Attachment with a content ID will
      unpack to a `(filename, payload)` tuple, but will compare unequally to
      that tuple.

      Thus, the following comparison will succeed:

          attachment = mail.Attachment('foo.jpg', 'data')
          filename, payload = attachment
          attachment == filename, payload


      ...while the following will fail:

          attachment = mail.Attachment('foo.jpg', 'data', content_id='<foo>')
          filename, payload = attachment
          attachment == filename, payload


      The following comparison will pass:

        attachment = mail.Attachment('foo.jpg', 'data', content_id='<foo>')
        attachment == (attachment.filename,
                       attachment.payload,
                       attachment.content_id)

  Attributes:
    filename: The name of the attachment.
    payload: The attachment data.
    content_id: Optional; the content ID for this attachment. Keyword only.
  """

  @_positional(3)
  def __init__(self, filename, payload, content_id=None):
    """Constructor.

    Arguments:
      filename: The name of the attachment.
      payload: The attachment data.
      content_id: Optional; the content ID for this attachment.
    """
    self.filename = filename
    self.payload = payload
    self.content_id = content_id

  def __eq__(self, other):
    self_tuple = (self.filename, self.payload, self.content_id)
    if isinstance(other, Attachment):
      other_tuple = (other.filename, other.payload, other.content_id)


    elif not hasattr(other, '__len__'):
      return NotImplemented
    elif len(other) == 2:
      other_tuple = other + (None,)
    elif len(other) == 3:
      other_tuple = other
    else:
      return NotImplemented
    return self_tuple == other_tuple

  def __hash__(self):
    if self.content_id:
      return hash((self.filename, self.payload, self.content_id))
    else:
      return hash((self.filename, self.payload))

  def __ne__(self, other):
    return not self == other

  def __iter__(self):
    return iter((self.filename, self.payload))

  def __getitem__(self, i):
    return tuple(iter(self))[i]

  def __contains__(self, val):
    return val in (self.filename, self.payload)

  def __len__(self):
    return 2


class EncodedPayload(object):
  """Wrapper for a payload that contains encoding information.

  When an email is received, it is usually encoded using a certain character
  set, then possibly further encoded using a transfer encoding in that
  character set. Most of the time, it is possible to decode the encoded
  payload as-is; however, in the case where it is not, the encoded payload and
  the original encoding information must be preserved.

  Attributes:
    payload: The original encoded payload.
    charset: The character set of the encoded payload. To specify that you want
        to use the default character set, set this argument to `None`.
    encoding: The transfer encoding of the encoded payload. To specify that you
        do not want the content to be encoded, set this argument to `None`.
  """

  def __init__(self, payload: bytes, charset=None, encoding=None):
    """Constructor.

    Args:
      payload: Maps to an attribute of the same name.
      charset: Maps to an attribute of the same name.
      encoding: Maps to an attribute of the same name.
    """
    if not isinstance(payload, bytes):
      raise TypeError('EncodedPayload must be bytes, but got ' + type(payload))

    self.payload = payload
    self.charset = charset
    self.encoding = encoding

  def decode(self):
    """Attempts to decode the encoded data.

    This function attempts to use `Python's codec library`_ to decode the
    payload. All exceptions are passed back to the caller.

    Returns:
      The binary or Unicode version of the payload content.

    .. _Python's codec library:
       https://docs.python.org/2/library/codecs.html
    """
    payload = self.payload


    if self.encoding and self.encoding.lower() not in ('7bit', '8bit'):
      try:
        payload = codecs.decode(payload, self.encoding)
      except LookupError:
        raise UnknownEncodingError('Unknown decoding %s.' % self.encoding)
      except (Exception, Error) as e:
        raise PayloadEncodingError('Could not decode payload: %s' % e)


    if self.charset and str(self.charset).lower() != '7bit':
      try:
        payload = codecs.decode(payload, str(self.charset))
      except LookupError:
        raise UnknownCharsetError('Unknown charset %s.' % self.charset)
      except (Exception, Error) as e:
        raise PayloadEncodingError('Could read characters: %s' % e)
    return payload

  def __eq__(self, other):
    """Equality operator.

    Args:
      other: The other `EncodedPayload` object with which to compare. Comparison
          with other object types are not implemented.

    Returns:
      `True` if the payload and encodings are equal; otherwise returns `False`.
    """
    if isinstance(other, EncodedPayload):
      return (self.payload == other.payload and
              self.charset == other.charset and
              self.encoding == other.encoding)
    else:
      return NotImplemented

  def __hash__(self):
    """Hashes an `EncodedPayload`."""
    return hash((self.payload, self.charset, self.encoding))

  def copy_to(self, mime_message):
    """Copies the contents of a message to a MIME message payload.

    If no content transfer encoding is specified and the character set does not
    equal the overall message encoding, the payload will be base64-encoded.

    Args:
      mime_message: Message instance that will receive the new payload.
    """
    if self.encoding:
      mime_message['content-transfer-encoding'] = self.encoding
    mime_message.set_payload(self.payload, self.charset)

  def to_mime_message(self):
    """Converts a message to a MIME message.

    Returns:
      The MIME message instance of the payload.
    """
    mime_message = email.message.Message()
    self.copy_to(mime_message)
    return mime_message

  def __str__(self):
    """String representation of an encoded message.

    Returns:
      The MIME encoded representation of an encoded payload as an independent
      message.
    """
    return str(self.to_mime_message())

  def __repr__(self):
    """Represents an encoded payload.

    Returns:
      The payload itself as represented by its hash value.
    """
    result = '<EncodedPayload payload=#%d' % hash(self.payload)
    if self.charset:
      result += ' charset=%s' % self.charset
    if self.encoding:
      result += ' encoding=%s' % self.encoding
    return result + '>'


class _EmailMessageBase(object):
  """Base class for Mail API service objects.

  Subclasses must define a class variable called `_API_CALL` with the name of
  its underlying mail-sending API call.
  """


  PROPERTIES = set([
      'sender',
      'reply_to',
      'subject',
      'body',
      'html',
      'amp_html',
      'attachments',
  ])

  ALLOWED_EMPTY_PROPERTIES = set([
      'subject',
      'body',
      'amp_html',
  ])



  PROPERTIES.update(('to', 'cc', 'bcc'))

  def __init__(self, mime_message=None, **kw):
    """Initializes an email message.

    This initializer creates a new `MailMessage` protocol buffer and initializes
    it with any keyword arguments.

    Args:
      mime_message: The MIME message to initialize from. If the message is an
          instance of `email.message.Message`, `mime_message` will take
          ownership as specified the original message.
      **kw: List of keyword properties as defined by `PROPERTIES`.
    """
    if mime_message:
      mime_message = _parse_mime_message(mime_message)
      self.update_from_mime_message(mime_message)
      self.__original = mime_message



    self.initialize(**kw)

  @property
  def original(self):
    """Gets the original MIME message from which values were set."""
    return self.__original

  def initialize(self, **kw):
    """Keyword initialization.

    This function sets all fields of the email message using the keyword
    arguments.

    Args:
      **kw: List of keyword properties as defined by `PROPERTIES`.
    """
    for name, value in six.iteritems(kw):
      setattr(self, name, value)


  def Initialize(self, **kw):
    self.initialize(**kw)

  def check_initialized(self):
    """Checks if `EmailMessage` is properly initialized.

    This function tests if `EmailMessage` meets the basic requirements to be
    used with the Mail API. To function properly, this function requires that
    the following fields must be set or have at least one value in the case of
    multi-value fields:
        - A subject must be set.
        - A recipient must be specified.
        - The message must contain a body.
        - All bodies and attachments must decode properly.

    This test does not include determining if the sender is actually authorized
    to send email for the application.

    Raises:
      InvalidAttachmentTypeError: The attachment type is invalid.
      MissingRecipientsError: No recipients were specified in the 'To', 'Cc', or
          'Bcc' fields.
      MissingSenderError: A sender was not specified.
      MissingSubjectError: A subject was not specified.
      MissingBodyError: A body was not specified.
      PayloadEncodingError: The payload was not properly encoded.
      UnknownEncodingError: The payload encoding could not be determined.
      UnknownCharsetError: The character set of the payload could not be
          determined.
    """
    if not hasattr(self, 'sender'):
      raise MissingSenderError()


    found_body = False

    try:
      body = self.body
    except AttributeError:
      pass
    else:
      if isinstance(body, EncodedPayload):

        body.decode()
      found_body = True

    try:
      html = self.html
    except AttributeError:
      pass
    else:
      if isinstance(html, EncodedPayload):

        html.decode()
      found_body = True

    try:
      amp_html = self.amp_html
    except AttributeError:
      pass
    else:
      if isinstance(amp_html, EncodedPayload):

        amp_html.decode()

    if hasattr(self, 'attachments'):
      for attachment in _attachment_sequence(self.attachments):


        _GetMimeType(attachment.filename)




        if isinstance(attachment.payload, EncodedPayload):
          attachment.payload.decode()


  def CheckInitialized(self):
    """Ensures that recipients have been specified."""
    self.check_initialized()

  def is_initialized(self):
    """Determines if `EmailMessage` is properly initialized.

    Returns:
      `True` if the message is properly initialized; otherwise returns `False`.
    """
    try:
      self.check_initialized()
      return True
    except Error:
      return False


  def IsInitialized(self):
    """Determines if `EmailMessage` is properly initialized."""
    return self.is_initialized()

  def ToProto(self):
    """Converts an email message to a protocol message.

    Unicode strings are converted to UTF-8 for all fields.

    This method is overridden by `EmailMessage` to support the sender field.

    Returns:
      The `MailMessage` protocol version of the mail message.

    Raises:
      Decoding errors when using `EncodedPayload` objects.
    """
    self.check_initialized()
    message = mail_service_pb2.MailMessage()
    message.Sender = _to_str(self.sender)

    if hasattr(self, 'reply_to'):
      message.ReplyTo = _to_str(self.reply_to)
    if hasattr(self, 'subject'):
      message.Subject = _to_str(self.subject)
    else:
      message.Subject = ''

    if hasattr(self, 'body'):
      body = self.body
      if isinstance(body, EncodedPayload):
        body = body.decode()
      message.TextBody = _to_str(body)

    if hasattr(self, 'html'):
      html = self.html
      if isinstance(html, EncodedPayload):
        html = html.decode()
      message.HtmlBody = _to_str(html)

    if hasattr(self, 'amp_html'):
      amp_html = self.amp_html
      if isinstance(amp_html, EncodedPayload):
        amp_html = amp_html.decode()
      message.AmpHtmlBody = _to_str(amp_html)

    if hasattr(self, 'attachments'):
      for attachment in _attachment_sequence(self.attachments):
        if isinstance(attachment.payload, EncodedPayload):
          attachment.payload = attachment.payload.decode()
        protoattachment = message.Attachment.add()
        protoattachment.FileName = _to_str(attachment.filename)
        protoattachment.Data = _to_str(attachment.payload)
        if attachment.content_id:
          protoattachment.ContentID = attachment.content_id
    return message

  def to_mime_message(self):
    """Generates a `MIMEMultipart` message from `EmailMessage`.

    This function calls `MailMessageToMessage` after converting `self` to
    the protocol buffer. The protocol buffer is better at handing corner cases
    than the `EmailMessage` class.

    Returns:
      A `MIMEMultipart` message that represents the provided `MailMessage`.

    Raises:
      InvalidAttachmentTypeError: The attachment type was invalid.
      MissingSenderError: A sender was not specified.
      MissingSubjectError: A subject was not specified.
      MissingBodyError: A body was not specified.
    """
    return mail_message_to_mime_message(self.ToProto())


  def ToMIMEMessage(self):
    return self.to_mime_message()

  def send(self, make_sync_call=apiproxy_stub_map.MakeSyncCall):
    """Sends an email message via the Mail API.

    Args:
      make_sync_call: Method that will make a synchronous call to the API proxy.
    """
    message = self.ToProto()
    response = api_base_pb2.VoidProto()

    try:
      make_sync_call('mail', self._API_CALL, message, response)
    except apiproxy_errors.ApplicationError as e:
      if e.application_error in ERROR_MAP:
        raise ERROR_MAP[e.application_error](e.error_detail)
      raise e


  def Send(self, *args, **kwds):
    self.send(*args, **kwds)

  def _check_attachment(self, attachment):

    if not (isinstance(attachment.filename, six.string_types) or
            isinstance(attachment.payload, six.string_types)):
      raise TypeError()

  def _check_attachments(self, attachments):
    """Checks the values that are going to the attachment field.

    This function is mainly used to check type safety of the values. Each value
    of the list must be a pair of the form `(file_name, data)`, and both values
    must use a string type.

    Args:
      attachments: Collection of attachment tuples.

    Raises:
      TypeError: If values do not use the string data type.
    """
    attachments = _attachment_sequence(attachments)
    for attachment in attachments:
      self._check_attachment(attachment)

  def __setattr__(self, attr, value):
    """Property that sets access control.

    This function controls write access to email fields.

    Args:
      attr: Attribute to access.
      value: The new value for the field.

    Raises:
      ValueError: If a value was not specified.
      AttributeError: If the attribute is not an allowed assignment field.
    """
    if not attr.startswith('_EmailMessageBase'):
      if attr in ['sender', 'reply_to']:
        check_email_valid(value, attr)

      if not value and attr not in self.ALLOWED_EMPTY_PROPERTIES:
        raise ValueError('May not set empty value for \'%s\'' % attr)


      if attr not in self.PROPERTIES:
        raise AttributeError('\'EmailMessage\' has no attribute \'%s\'' % attr)


      if attr == 'attachments':
        self._check_attachments(value)

    super(_EmailMessageBase, self).__setattr__(attr, value)

  def _add_body(self, content_type, payload):
    """Adds a body to an email from the payload.

    This function will overwrite any existing default plain or HTML body.

    Args:
      content_type: Content type of the body.
      payload: Payload to store the body as.
    """

    if content_type == 'text/plain':
      self.body = payload
    elif content_type == 'text/html':
      self.html = payload

  def _update_payload(self, mime_message: email.message.Message):
    """Updates a payload of a mail message from `mime_message`.

    This function works recursively when it receives a multi-part body. If the
    function receives a non-multi-part MIME object, it will determine whether
    it is an attachment by whether it contains a file name. Attachments and
    bodies are then wrapped in `EncodedPayload` that use the correct character
    sets and encodings.

    Args:
      mime_message: A Message MIME email object.
    """
    payload = mime_message.get_payload()

    if payload:
      if mime_message.is_multipart():
        for alternative in payload:
          self._update_payload(alternative)
      else:
        filename = mime_message.get_param('filename',
                                          header='content-disposition')
        if filename:

          filename = email.utils.collapse_rfc2231_value(filename)
        if not filename:
          filename = mime_message.get_param('name')


        encoded_payload = EncodedPayload(
            six.ensure_binary(payload),
            (mime_message.get_content_charset() or mime_message.get_charset()),
            mime_message['content-transfer-encoding'])

        if 'content-id' in mime_message:
          attachment = Attachment(
              filename, encoded_payload, content_id=mime_message['content-id'])
        else:
          attachment = Attachment(filename, encoded_payload)

        if filename:

          try:
            attachments = self.attachments
          except AttributeError:
            self.attachments = [attachment]
          else:
            if isinstance(attachments[0], six.string_types):
              self.attachments = [attachments]
              attachments = self.attachments
            attachments.append(attachment)
        else:
          self._add_body(mime_message.get_content_type(), encoded_payload)

  def update_from_mime_message(self, mime_message):
    """Copies information from a MIME message.

    Information from the `email.Message` instance is set to the values of the
    MIME message. This function will only copy values that it finds. Missing
    values will not be copied, nor will those values overwrite old values with
    blank values.

    This object is not guaranteed to be initialized after this call.

    Args:
      mime_message: The `email.Message` instance from which to copy information.

    Returns:
      A MIME message instance of the `mime_message`.
    """
    mime_message = _parse_mime_message(mime_message)

    sender = _decode_and_join_header(mime_message['from'])
    if sender:
      self.sender = sender

    reply_to = _decode_and_join_header(mime_message['reply-to'])
    if reply_to:
      self.reply_to = reply_to

    subject = _decode_and_join_header(mime_message['subject'], separator=u'')
    if subject:
      self.subject = subject

    self._update_payload(mime_message)

  def bodies(self, content_type: typing.Text = None):
    """Iterates over all bodies.

    Args:
      content_type: Content type on which to filter. This argument allows you to
          select only specific types of content. You can use the base type or
          the content type.

          For example:

              content_type = 'text/html'  # Matches only HTML content.
              content_type = 'text'       # Matches text of any kind.

    Yields:
      A `(content_type, payload)` tuple for HTML and body in that order.
    """
    if (not content_type or
        content_type == 'text' or
        content_type == 'text/html'):
      try:
        yield 'text/html', self.html
      except AttributeError:
        pass

    if (not content_type or
        content_type == 'text' or
        content_type == 'text/plain'):
      try:
        yield 'text/plain', self.body
      except AttributeError:
        pass


class EmailMessage(_EmailMessageBase):
  """Main interface to the Mail API service.

  This class is used to programmatically build an email message to send via the
  Mail API. To use this class, construct an instance, populate its fields and
  call `Send()`.

  An `EmailMessage` can be built completely by the constructor:

      EmailMessage(sender='sender@nowhere.com',
                   to='recipient@nowhere.com',
                   subject='a subject',
                   body='This is an email to you').Send()


  You might want your application to build an email in different places
  throughout the code. For this usage, `EmailMessage` is mutable:

        message = EmailMessage()
        message.sender = 'sender@nowhere.com'
        message.to = ['recipient1@nowhere.com', 'recipient2@nowhere.com']
        message.subject = 'a subject'
        message.body = 'This is an email to you'
        message.check_initialized()
        message.send()
  """

  _API_CALL = 'Send'
  PROPERTIES = set(_EmailMessageBase.PROPERTIES | set(('headers',)))


  def __init__(
      self,
      mime_message: typing.Optional[typing.Union[typing.Text, bytes,
                                                 typing.TextIO,
                                                 email.message.Message]] = None,
      **kw):
    """
    """


    super().__init__(mime_message, **kw)


  def check_initialized(self):
    """Provides additional checks to ensure that recipients have been specified.

    Raises:
      MissingRecipientError: If no recipients are specified in 'To', 'Cc', or
          'Bcc'.
    """
    if (not hasattr(self, 'to') and
        not hasattr(self, 'cc') and
        not hasattr(self, 'bcc')):
      raise MissingRecipientsError()
    super(EmailMessage, self).check_initialized()


  def CheckInitialized(self):
    """Ensures that recipients have been specified."""
    self.check_initialized()

  def ToProto(self):
    """Performs more conversion of recipient fields to the protocol buffer.

    Returns:
      The `MailMessage` protocol version of the mail message, including sender
      fields.
    """
    message = super(EmailMessage, self).ToProto()

    for attribute, adder in (('to', message.To.append),
                             ('cc', message.Cc.append), ('bcc',
                                                         message.Bcc.append)):
      if hasattr(self, attribute):
        for address in _email_sequence(getattr(self, attribute)):
          adder(_to_str(address))
    for name, value in six.iteritems(getattr(self, 'headers', {})):
      header = message.Header.add()
      header.name = name
      header.value = _to_str(value)
    return message

  def __setattr__(self, attr, value):
    """Performs additional checks on recipient fields."""

    if attr in ['to', 'cc', 'bcc']:
      if isinstance(value, six.string_types):
        if value == '' and getattr(self, 'ALLOW_BLANK_EMAIL', False):
          return
        check_email_valid(value, attr)
      else:
        for address in value:
          check_email_valid(address, attr)
    elif attr == 'headers':
      check_headers_valid(value)

    super(EmailMessage, self).__setattr__(attr, value)

  def update_from_mime_message(self, mime_message):
    """Copies information for recipients from a MIME message.

    Args:
      mime_message: The `email.Message` instance from which to copy information.
    """
    mime_message = _parse_mime_message(mime_message)
    super(EmailMessage, self).update_from_mime_message(mime_message)

    to = _decode_address_list_field(mime_message.get_all('to'))
    if to:
      self.to = to

    cc = _decode_address_list_field(mime_message.get_all('cc'))
    if cc:
      self.cc = cc

    bcc = _decode_address_list_field(mime_message.get_all('bcc'))
    if bcc:
      self.bcc = bcc


class AdminEmailMessage(_EmailMessageBase):
  """Interface that sends email messages to all administrators via the Mail API.

  This class is used to programmatically build an administrator email message to
  be sent via the Mail API. To use the class, construct an instance, populate
  its fields, and call `Send()`.

  Unlike normal email messages, addresses in the recipient fields are ignored
  and not used to send the message.

  An `AdminEmailMessage` can be built completely by the constructor:

      AdminEmailMessage(sender='sender@nowhere.com',
                        subject='a subject',
                        body='This is an email to you').Send()


  You might want your application to build an administrator email in different
  places throughout the code. For this, `AdminEmailMessage` is mutable:

      message = AdminEmailMessage()
      message.sender = 'sender@nowhere.com'
      message.subject = 'a subject'
      message.body = 'This is an email to you'
      message.check_initialized()
      message.send()
  """

  _API_CALL = 'SendToAdmins'
  __UNUSED_PROPERTIES = set(('to', 'cc', 'bcc'))

  def __setattr__(self, attr, value):
    if attr in self.__UNUSED_PROPERTIES:
      logging.warning('\'%s\' is not a valid property to set '
                      'for AdminEmailMessage.  It is unused.', attr)
    super(AdminEmailMessage, self).__setattr__(attr, value)


class InboundEmailMessage(EmailMessage):
  """Receives a parsed email as it is recevied from an external source.

  This class makes use of a `date` field and can store any number of additional
  bodies. These additional attributes make the email more flexible as required
  for incoming mail, where the developer has less control over the content.

  Example:

      # Read mail message from `CGI` input.
      message = InboundEmailMessage(sys.stdin.read())
      logging.info('Received email message from %s at %s',
                   message.sender,
                   message.date)
      enriched_body = list(message.bodies('text/enriched'))[0]
      # ... Do something with body ...
  """

  __HEADER_PROPERTIES = {
      'date': 'date',
      'message_id': 'message-id',
  }

  PROPERTIES = frozenset(_EmailMessageBase.PROPERTIES
                         | set(('alternate_bodies',))
                         | set(six.iterkeys(__HEADER_PROPERTIES)))

  ALLOW_BLANK_EMAIL = True

  @classmethod
  def from_environ(cls, environ):
    """Creates an email message by parsing the HTTP request body in `environ`.

    Example (WSGI)::

      def app(environ, start_response):
        mail_message = mail.InboundEmailMessage.from_environ(request.environ)

        # Do something with the message
        logging.info('Received greeting from %s: %s' % (mail_message.sender,
                                                          mail_message.body))
        start_response("200 OK", [])
        return “Success”

    Note: Flask (other web frameworks) can directly use
          `new InboundEmailMessage(request_bytes)` to create the email message
          if they have the request bytes of an HTTP request.

    Args:
      environ: a WSGI dict describing the HTTP request (See PEP 333).
    Returns:
      An InboundEmailMessage object.
    """
    try:
      req_size = int(environ.get('CONTENT_LENGTH', 0))
    except ValueError:
      req_size = 0

    request_bytes = environ['wsgi.input'].read(req_size)
    return InboundEmailMessage(request_bytes)

  def update_from_mime_message(self, mime_message):
    """Updates the values of a MIME message.

    This function copies over date values.

    Args:
      mime_message: The `email.Message` instance from which you want to copy
        information.
    """
    mime_message = _parse_mime_message(mime_message)
    super(InboundEmailMessage, self).update_from_mime_message(mime_message)

    for property_, header in six.iteritems(
        InboundEmailMessage.__HEADER_PROPERTIES):
      value = mime_message[header]
      if value:
        setattr(self, property_, value)

  def _add_body(self, content_type, payload):
    """Adds a body to an inbound message.

    This method is overidden to handle incoming messages that have more than one
    plain or HTML body or have any unidentified bodies.

    This method will not overwrite existing HTML and body values. Therefore,
    when the message is updated with the body, the text and HTML bodies that
    are first in the MIME document order are assigned to the body and HTML
    properties.

    Args:
      content_type: Content type of the additional body.
      payload: Content of the additional body.
    """
    if (content_type == 'text/plain' and not hasattr(self, 'body') or
        content_type == 'text/html' and not hasattr(self, 'html')):

      super(InboundEmailMessage, self)._add_body(content_type, payload)
    else:

      try:
        alternate_bodies = self.alternate_bodies
      except AttributeError:
        alternate_bodies = self.alternate_bodies = [(content_type, payload)]
      else:
        alternate_bodies.append((content_type, payload))

  def bodies(self, content_type: typing.Text = None):
    """Iterates over all bodies.

    Args:
      content_type: Content type on which to filter. This argument allows you to
          select only specific types of content. You can use the base type or
          the content type.

          For example, `content_type = 'text/html'` matches only HTML content,
          and `content_type = 'text'` matches text of any kind.


    Yields:
      A `(content_type, payload)` tuple for all bodies of a message, including
      the body, HTML, and all `alternate_bodies`, in that order.
    """
    main_bodies = super(InboundEmailMessage, self).bodies(content_type)
    for payload_type, payload in main_bodies:
      yield payload_type, payload

    partial_type = bool(content_type and content_type.find('/') < 0)

    try:
      for payload_type, payload in self.alternate_bodies:
        if content_type:
          if partial_type:
            match_type = payload_type.split('/')[0]
          else:
            match_type = payload_type
          match = match_type == content_type
        else:
          match = True

        if match:
          yield payload_type, payload
    except AttributeError:
      pass

  def to_mime_message(self):
    """Converts a message to a MIME message.

    This function adds additional headers from the inbound email.

    Returns:
      The MIME message instance of a payload.
    """
    mime_message = super(InboundEmailMessage, self).to_mime_message()

    for property_, header in six.iteritems(
        InboundEmailMessage.__HEADER_PROPERTIES):
      try:
        mime_message[header] = getattr(self, property_)
      except AttributeError:
        pass

    return mime_message










parser.Parser



class _MultiDict(MutableMapping):
  """A slim version of the WebOb.MultiDict class.

  This only includes functionality needed for accessing POST form vars
  needed for for parsing a BounceNotification object.
  Original WebOb class:
  https://github.com/Pylons/webob/blob/master/src/webob/multidict.py
  """

  def __init__(self):
    self._items = []

  def __len__(self):
    return len(self._items)

  def __getitem__(self, key):
    for k, v in reversed(self._items):
      if k == key:
        return v
    raise KeyError(key)

  def __setitem__(self, key, value):
    try:
      del self[key]
    except KeyError:
      pass
    self._items.append((key, value))

  def __delitem__(self, key):
    items = self._items
    found = False

    for i in range(len(items) - 1, -1, -1):
      if items[i][0] == key:
        del items[i]
        found = True

    if not found:
      raise KeyError(key)

  def add(self, key, value):
    """Add the key and value, not overwriting any previous value."""
    self._items.append((key, value))

  def keys(self):
    for k, _ in self._items:
      yield k

  __iter__ = keys

  @classmethod
  def from_fieldstorage(cls, fs):
    """Create a MultiDict from a cgi.FieldStorage instance.

    This mimics functionality in webob.MultiDict without taking a dependency
    on the WebOb package -
    https://github.com/Pylons/webob/blob/259230aa2b8b9cf675c996e157c5cf021c256059/src/webob/multidict.py#L57

    Args:
      fs: cgi.FieldStorage object correspoinding to a POST request
    Returns:
      MultiDict that contains form variables from the POST request
    """
    obj = cls()


    for field in fs.list or ():
      charset = field.type_options.get('charset', 'utf8')
      transfer_encoding = field.headers.get('Content-Transfer-Encoding', None)
      supported_transfer_encoding = {
          'base64': binascii.a2b_base64,
          'quoted-printable': binascii.a2b_qp,
      }

      if charset == 'utf8':
        def decode(b):
          return b
      else:
        def decode(b):
          return b.encode('utf8').decode(charset)

      if field.filename:
        field.filename = decode(field.filename)
        obj.add(field.name, field)
      else:
        value = field.value

        if transfer_encoding in supported_transfer_encoding:

          value = value.encode('utf8')
          value = supported_transfer_encoding[transfer_encoding](value)


          value = value.decode('utf8')
        obj.add(field.name, decode(value))

    return obj


class BounceNotification(object):
  """Encapsulates a bounce notification received by the application."""

  def __init__(self, post_vars: typing.Mapping[str, typing.Any]):
    """Constructs a new BounceNotification from an HTTP request.

    Properties:
      original: a dict describing the message that caused the bounce.
      notification: a dict describing the bounce itself.
      original_raw_message: the raw message that caused the bounce.

    The 'original' and 'notification' dicts contain the following keys:
      to, cc, bcc, from, subject, text

    Args:
      post_vars: a dictionary with keys as strings. This should
        contain bounce information, and the following keys are handled:
          original-from
          original-to
          original-cc
          original-bcc
          original-subject
          original-text
          notification-from
          notification-to
          notification-cc
          notification-bcc
          notification-subject
          notification-text
          raw-message
        For all keys except 'raw-message', the value can be anything.
        The Bounce Notification object just assigns these values to the
        `original` and `notification` properties of this instance,
        which are dictionaries.
        For example, original["to"] = post_vars.get("original-to")

        The `raw-message` value is used to create an `InboundEmailMessage`
        object. This value should be a valid input to the `EmailMessage`
        constructor (inherited by 'InboundEmailMessage').

        Flask- This is typically the `flask.request.form` field (if
        the user wants to pass single (non-list) values for each key), or
        'dict(flask.request.form.lists())' if the user wants to store a list for
        each key (example use case is multiple `to` and `cc` recipients).

        Webob- `webob.Request.POST` can be used for single values, and
        `webob.Request.POST.dict_of_lists()` for multiple values.

        Django- `request.POST` can be used for single values, and
        `dict(request.POST.lists())` can be used for multiple values.
    """
    self.__original = {}
    self.__notification = {}
    for field in ['to', 'cc', 'bcc', 'from', 'subject', 'text']:
      self.__original[field] = post_vars.get('original-' + field, '')
      self.__notification[field] = post_vars.get('notification-' + field, '')

    raw_message = post_vars.get('raw-message', '')

    if isinstance(raw_message, list):
      if not raw_message:

        raw_message = ''
      elif len(raw_message) > 1:
        raise ValueError('Multiple values found for "raw-message" in post_vars.'
                         ' Expected single value.')
      else:
        raw_message = raw_message[0]

      raw_message = raw_message[0] if raw_message else ''
    self.__original_raw_message = InboundEmailMessage(raw_message)

  @classmethod
  def from_environ(cls, environ):
    """Transforms the HTTP request body to a bounce notification object.

    Example(WSGI)::

      def BounceReceiver(environ, start_response):
        bounce_msg = mail.BounceNotification.from_environ(environ)

        # Add logic for what to do with the bounce notification
        print('Bounce original: %s', bounce_msg.original)
        print('Bounce notification: %s', bounce_msg.notification)

        # Return suitable response
        response = http.HTTPStatus.OK
        start_response(f'{response.value} {response.phrase}', [])
        return ['success'.encode('utf-8')]

    Args:
      environ: a WSGI dict describing the HTTP request (See PEP 333).
    Returns:
      A BounceNotification object.
    """
    fs = cgi.FieldStorage(
        fp=environ['wsgi.input'],
        environ=environ,
        keep_blank_values=True,
        encoding='utf8',
    )

    post_vars = _MultiDict.from_fieldstorage(fs)
    return BounceNotification(post_vars)

  @property
  def original(self):
    return self.__original

  @property
  def notification(self):
    return self.__notification

  @property
  def original_raw_message(self):
    return self.__original_raw_message



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""A Python blobstore API used by app developers.

This module contains methods that are used to interface with Blobstore API.
The module includes a `db.Model`-like class that represents a reference to a
very large blob. The module imports a `db.Key`-like class that represents a blob
key.
"""











import base64
import cgi
import collections
import email
import email.message
import http
import re

from google.appengine.api import datastore
from google.appengine.api import datastore_errors
from google.appengine.api import datastore_types
from google.appengine.api.blobstore import blobstore
from google.appengine.ext import db
import six
from six.moves import urllib

__all__ = ['BLOB_INFO_KIND',
           'BLOB_KEY_HEADER',
           'BLOB_MIGRATION_KIND',
           'BLOB_RANGE_HEADER',
           'BlobFetchSizeTooLargeError',
           'BlobInfo',
           'BlobInfoParseError',
           'BlobKey',
           'BlobMigrationRecord',
           'BlobNotFoundError',
           'BlobReferenceProperty',
           'BlobReader',
           'FileInfo',
           'FileInfoParseError',
           'DataIndexOutOfRangeError',
           'PermissionDeniedError',
           'Error',
           'InternalError',
           'MAX_BLOB_FETCH_SIZE',
           'UPLOAD_INFO_CREATION_HEADER',
           'CLOUD_STORAGE_OBJECT_HEADER',
           'create_rpc',
           'create_upload_url',
           'create_upload_url_async',
           'delete',
           'delete_async',
           'fetch_data',
           'fetch_data_async',
           'create_gs_key',
           'create_gs_key_async',
           'GS_PREFIX',
           'get',
           'parse_blob_info',
           'parse_file_info',
           'RangeFormatError',
           'UnsupportedRangeFormatError',
           'BlobstoreDownloadHandler',
           'BlobstoreUploadHandler',]

Error = blobstore.Error
InternalError = blobstore.InternalError
BlobFetchSizeTooLargeError = blobstore.BlobFetchSizeTooLargeError
BlobNotFoundError = blobstore.BlobNotFoundError
_CreationFormatError = blobstore._CreationFormatError
DataIndexOutOfRangeError = blobstore.DataIndexOutOfRangeError
PermissionDeniedError = blobstore.PermissionDeniedError

BlobKey = blobstore.BlobKey
create_rpc = blobstore.create_rpc
create_upload_url = blobstore.create_upload_url
create_upload_url_async = blobstore.create_upload_url_async
delete = blobstore.delete
delete_async = blobstore.delete_async
create_gs_key = blobstore.create_gs_key
create_gs_key_async = blobstore.create_gs_key_async


BLOB_INFO_KIND = blobstore.BLOB_INFO_KIND
BLOB_MIGRATION_KIND = blobstore.BLOB_MIGRATION_KIND
BLOB_KEY_HEADER = blobstore.BLOB_KEY_HEADER
BLOB_RANGE_HEADER = blobstore.BLOB_RANGE_HEADER
MAX_BLOB_FETCH_SIZE = blobstore.MAX_BLOB_FETCH_SIZE
UPLOAD_INFO_CREATION_HEADER = blobstore.UPLOAD_INFO_CREATION_HEADER
CLOUD_STORAGE_OBJECT_HEADER = blobstore.CLOUD_STORAGE_OBJECT_HEADER
GS_PREFIX = blobstore.GS_PREFIX

_CONTENT_DISPOSITION_FORMAT = b'attachment; filename="%s"'
_CONTENT_DISPOSITION_FORMAT_UTF8 = (b'attachment; filename="%s"; '
                                    b'filename*=utf-8\'\'%s')

_SEND_BLOB_PARAMETERS = frozenset(['use_range'])

_RANGE_NUMERIC_FORMAT = r'([0-9]*)-([0-9]*)'
_RANGE_FORMAT = r'([a-zA-Z]+)=%s' % _RANGE_NUMERIC_FORMAT
_RANGE_FORMAT_REGEX = re.compile('^%s$' % _RANGE_FORMAT)
_UNSUPPORTED_RANGE_FORMAT_REGEX = re.compile(
    '^%s(?:,%s)+$' % (_RANGE_FORMAT, _RANGE_NUMERIC_FORMAT))
_BYTES_UNIT = 'bytes'


class BlobInfoParseError(Error):
  """The CGI parameter does not contain a valid `BlobInfo` record."""


class FileInfoParseError(Error):
  """The CGI parameter does not contain a valid `FileInfo` record."""


class RangeFormatError(Error):
  """Raised when Range header incorrectly formatted."""


class UnsupportedRangeFormatError(RangeFormatError):
  """Raised when Range format is correct, but not supported."""




class _GqlQuery(db.GqlQuery):
  """GqlQuery class that explicitly sets `model-class`.

  This does the same as the original `db.GqlQuery` class except that it does
  not try to find the model class based on the compiled GQL query. The
  caller instead provides the query with a model class to use for construction.

  Note:
      This class is required for compatibility with the current `db.py` query
      mechanism but will be removed in the future. DO NOT USE.
  """


  def __init__(self, query_string, model_class, *args, **kwds):
    """Constructor.

    Args:
      query_string: A properly formatted GQL query string.
      model_class: The model class from which entities are constructed.
      *args: Positional arguments that are used to bind numeric references in
          the query.
      **kwds: Dictionary-based arguments for named references.
    """


    from google.appengine.ext import gql
    app = kwds.pop('_app', None)
    self._proto_query = gql.GQL(query_string, _app=app, namespace='')

    super(db.GqlQuery, self).__init__(model_class)
    self.bind(*args, **kwds)




class BlobInfo(object):
  """Information about blobs in Blobstore.

  This is a `db.Model`-like class that contains information about blobs that are
  stored by an application. Like `db.Model`, this class is backed by a
  Datastore entity; however, `BlobInfo` instances are read-only and have a much
  more limited interface.

  Each `BlobInfo` has a key of type `BlobKey` associated with it. This key is
  specific to the Blobstore API and is not compatible with `db.get`. The key
  can be used for quick lookup by passing it to `BlobInfo.get`. This key
  converts easily to a string, which is web safe and can be embedded in URLs.

  Properties:
      - content_type: The content type of the blob.
      - creation: The creation date of the blob, or when it was uploaded.
      - filename: The file name that the user selected from their machine.
      - size: The size of the uncompressed blob.
      - md5_hash: The MD5 hash value of the uploaded blob.
    gs_object_name: The name of the object, if the blob is stored in
      Google Cloud Storage, in the form /[bucket-name]/[object-name]

  All properties are read-only. Attempting to assign a value to a property
  will raise a `NotImplementedError`.
  """

  _unindexed_properties = frozenset([])


  _all_properties = frozenset(['content_type', 'creation', 'filename',
                               'size', 'md5_hash', 'gs_object_name'])

  @property
  def content_type(self):
    """Returns the content type of the blob.

    Returns:
      The content type of the blob.
    """
    return self.__get_value('content_type')

  @property
  def creation(self):
    """Returns the creation date or time of upload of the blob.

    Returns:
      The creation date or time of upload of the blob.
    """
    return self.__get_value('creation')

  @property
  def filename(self):
    """Returns the file name that the user selected from their machine.

    Returns:
      The file name that the user selected.
    """
    return self.__get_value('filename')

  @property
  def size(self):
    """Returns the size of the uncompressed blob.

    Returns:
      The size of the uncompressed blob.
    """
    return self.__get_value('size')

  @property
  def md5_hash(self):
    """Returns the MD5 hash value of the uncompressed blob.

    Returns:
      The hash value of the uncompressed blob.
    """
    return self.__get_value('md5_hash')

  @property
  def gs_object_name(self):
    return self.__get_value('gs_object_name')

  def __init__(self, entity_or_blob_key, _values=None):
    """Constructor for wrapping the blobstore entity.

    The constructor should not be used outside of this package and tests.

    Args:
      entity_or_blob_key: The datastore entity or blob key that represents the
          blob reference.
      _values: Optional; not recommended. This argument passes the associated
          entity when `entity_or_blob_key` is a blob key.
    """
    if isinstance(entity_or_blob_key, datastore.Entity):
      self.__entity = entity_or_blob_key
      self.__key = BlobKey(entity_or_blob_key.key().name())
    elif isinstance(entity_or_blob_key, BlobKey):
      self.__entity = _values
      self.__key = entity_or_blob_key
    else:
      raise TypeError('Must provide Entity or BlobKey')



  @classmethod
  def from_entity(cls, entity):
    """Converts an entity to `BlobInfo`.

    Note:
        This method is required for compatibility with the current `db.py` query
        mechanism but will be removed in the future. DO NOT USE.

    Args:
      entity: The entity that you are trying to convert.

    Returns:
      The `BlobInfo` that was converted from the entity.
    """
    return BlobInfo(entity)



  @classmethod
  def properties(cls):
    """Defines the set of properties that belong to `BlobInfo`.

    Note:
        This method is required for compatibility with the current `db.py` query
        mechanism but will be removed in the future. DO NOT USE.

    Returns:
      A set of all of the properties that belong to `BlobInfo`.
    """
    return set(cls._all_properties)

  def __get_value(self, name):
    """Gets a `BlobInfo` value, loading an entity if necessary.

    This method allows lazy loading of the underlying datastore entity.  It
    should never be invoked directly.

    Args:
      name: The name of property for which you want to retrieve the value.

    Returns:
      The value of the `BlobInfo` property from the entity.
    """
    if self.__entity is None:
      self.__entity = datastore.Get(
          datastore_types.Key.from_path(
              self.kind(), str(self.__key), namespace=''))
    return self.__entity.get(name, None)

  def key(self):
    """Gets the key for a blob.

    Returns:
      The `BlobKey` instance that identifies this blob.
    """
    return self.__key

  def delete(self, _token=None):
    """Permanently deletes a blob from Blobstore."""
    delete(self.key(), _token=_token)

  def open(self, *args, **kwargs):
    """Returns a `BlobReader` for this blob.

    Args:
      *args: Arguments to pass to the `BlobReader` constructor.
      **kwargs: Keyword arguments to pass to the `BlobReader` constructor.
    Returns:
      A `BlobReader` instance.
    """
    return BlobReader(self, *args, **kwargs)

  @classmethod
  def get(cls, blob_keys):
    """Retrieves a `BlobInfo` by key or by a list of keys.

    Args:
      blob_keys: A key or a list of keys. Keys can be in string, Unicode, or
          `BlobKey` format.

    Returns:
      A `BlobInfo` instance that is associated with the provided key or a list
      of `BlobInfo` instances if a list of keys was provided. Keys that are not
      found in Blobstore return `None`.
    """
    blob_keys = cls.__normalize_and_convert_keys(blob_keys)
    try:
      entities = datastore.Get(blob_keys)
    except datastore_errors.EntityNotFoundError:
      return None
    if isinstance(entities, datastore.Entity):
      return BlobInfo(entities)
    else:
      references = []
      for entity in entities:
        if entity is not None:
          references.append(BlobInfo(entity))
        else:
          references.append(None)
      return references

  @classmethod
  def all(cls):
    """Creates a query for all `BlobInfo` objects associated with the app.

    Returns:
      A `db.Query` object that queries over `BlobInfo`'s datastore kind.
    """
    return db.Query(model_class=cls, namespace='')

  @classmethod
  def __factory_for_kind(cls, kind):
    if kind == BLOB_INFO_KIND:
      return BlobInfo
    raise ValueError('Cannot query for kind %s' % kind)

  @classmethod
  def gql(cls, query_string, *args, **kwds):
    """Returns a query using a GQL query string.

    See the `GQL source` for more information about GQL.

    Args:
      query_string: A properly formatted GQL query string that omits
          `SELECT * FROM <entity>`
      *args: The remaining positional arguments that are used to bind numeric
          references in the query.
      **kwds: The dictionary-based arguments for named parameters.

    Returns:
      A `gql.GqlQuery` object that queries over `BlobInfo`'s datastore kind.

    .. _GQL source:
       https://cloud.google.com/appengine/docs/python/refdocs/google.appengine.ext.gql
    """
    return _GqlQuery('SELECT * FROM %s %s'
                       % (cls.kind(), query_string),
                     cls,
                     *args,
                     **kwds)


  @classmethod
  def kind(self):
    """Gets the entity kind for the `BlobInfo`.

    Note:
        This method is required for compatibility with the current `db.py` query
        mechanism but will be removed in the future. DO NOT USE.

    Returns:
      The entity kind for `BlobInfo`.
    """
    return BLOB_INFO_KIND

  @classmethod
  def __normalize_and_convert_keys(cls, keys):
    """Normalizes and converts all keys to `BlobKey` type.

    This method is based on `datastore.NormalizeAndTypeCheck()`.

    Args:
      keys: A single key or a list or tuple of keys. Keys can be a string or
          a `BlobKey`.

    Returns:
      A single key or a list with all of the strings replaced by `BlobKey`
      instances.
    """
    if isinstance(keys, (list, tuple)):
      multiple = True

      keys = list(keys)
    else:
      multiple = False
      keys = [keys]

    for index, key in enumerate(keys):
      if not isinstance(key, (six.string_types, BlobKey)):
        raise datastore_errors.BadArgumentError(
            'Expected string or BlobKey; received %s (a %s)' % (
                key,
                datastore.typename(key)))
      keys[index] = datastore.Key.from_path(cls.kind(), str(key), namespace='')

    if multiple:
      return keys
    else:
      return keys[0]


def get(blob_key):
  """Gets a `BlobInfo` record from blobstore.

  Does the same as `BlobInfo.get`.

  Args:
    blob_key: The `BlobKey` of the record you want to retrieve.

  Returns:
      A `BlobInfo` instance that is associated with the provided key or a list
      of `BlobInfo` instances if a list of keys was provided. Keys that are not
      found in Blobstore return `None`.
  """
  return BlobInfo.get(blob_key)


def _get_upload_content(field_storage):
  """Returns an `email.Message` that contains the values of the file transfer.

  This method decodes the content of the field storage and creates a new
  `email.Message`.

  Args:
    field_storage: `cgi.FieldStorage` that represents an uploaded blob.

  Returns:
    An `email.message.Message` that lists the upload information.
  """
  message = email.message.Message()
  message.add_header(
      'content-transfer-encoding',
      field_storage.headers.get('Content-Transfer-Encoding', ''))
  message.set_payload(field_storage.file.read())
  payload = message.get_payload(decode=True)
  return email.message_from_string(payload.decode('utf-8'))


def _parse_upload_info(field_storage, error_class):
  """Parses the upload information from the file upload `field_storage`.

  Args:
    field_storage: `cgi.FieldStorage` that represents the uploaded blob.
    error_class: The error to raise if the information cannot be parsed.

  Returns:
    A dictionary that contains the parsed values. This method will return `None`
    if `field_storage` was not defined.

  Raises:
    `error_class` when provided with a `field_storage` that does not contain
    enough information.
  """
  if field_storage is None:
    return None

  field_name = field_storage.name

  def get_value(dict, name):
    """Gets the name of the field."""
    value = dict.get(name, None)
    if value is None:
      raise error_class(
          'Field %s has no %s.' % (field_name, name))
    return value

  filename = get_value(field_storage.disposition_options, 'filename')
  blob_key = field_storage.type_options.get('blob-key', None)

  upload_content = _get_upload_content(field_storage)


  field_storage.file.seek(0)
  content_type = get_value(upload_content, 'content-type')
  size = get_value(upload_content, 'content-length')
  creation_string = get_value(upload_content, UPLOAD_INFO_CREATION_HEADER)
  md5_hash_encoded = get_value(upload_content, 'content-md5')
  md5_hash = base64.urlsafe_b64decode(md5_hash_encoded)
  gs_object_name = upload_content.get(CLOUD_STORAGE_OBJECT_HEADER, None)

  try:
    size = int(size)
  except (TypeError, ValueError):
    raise error_class(
        '%s is not a valid value for %s size.' % (size, field_name))

  try:
    creation = blobstore._parse_creation(creation_string, field_name)
  except blobstore._CreationFormatError as err:
    raise error_class(str(err))

  return {'blob_key': blob_key,
          'content_type': content_type,
          'creation': creation,
          'filename': filename,
          'size': size,
          'md5_hash': md5_hash,
          'gs_object_name': gs_object_name,
         }


def parse_blob_info(field_storage):
  """Parses a `BlobInfo` record from file upload `field_storage`.

  Args:
    field_storage: `cgi.FieldStorage` that represents an uploaded blob.

  Returns:
    A `BlobInfo` record as parsed from the `field_storage` instance. This
    method will return `None` if `field_storage` was not defined.

  Raises:
    BlobInfoParseError: If the provided `field_storage` does not contain enough
        information to construct a `BlobInfo` object.
  """
  info = _parse_upload_info(field_storage, BlobInfoParseError)

  if info is None:
    return None

  key = info.pop('blob_key', None)
  if not key:
    raise BlobInfoParseError('Field %s has no %s.' % (field_storage.name,
                                                      'blob_key'))

  return BlobInfo(BlobKey(key), info)


class FileInfo(object):
  """Details information about uploaded files.

  This class contains information about blobs stored by an application.

  This class is similar to `BlobInfo`; however, this method does not make use of
  a key, and the information is not persisted in the datastore.

  Properties:
      - content_type: The content type of the uploaded file.
      - creation: The creation date of the uploaded file, or when it was
          uploaded.
      - filename: The file name that the user selected from their machine.
      - size: The size of the uncompressed file.
      - md5_hash: The MD5 hash value of the uploaded file.
      - gs_object_name: The name of the file that was written to Google Cloud
        Storage, or `None` if the file was not uploaded to Google Cloud Storage.

  All properties are read-only. Attempting to assign a value to a property
  will raise an `AttributeError`.
  """

  def __init__(self, filename=None, content_type=None, creation=None,
               size=None, md5_hash=None, gs_object_name=None):
    self.__filename = filename
    self.__content_type = content_type
    self.__creation = creation
    self.__size = size
    self.__md5_hash = md5_hash
    self.__gs_object_name = gs_object_name

  @property
  def filename(self):
    """Returns the file name that the user selected.

    Returns:
      The file name that the user selected.
    """
    return self.__filename

  @property
  def content_type(self):
    """Returns the content type of the uploaded file.

    Returns:
      The content type of the file.
    """
    return self.__content_type

  @property
  def creation(self):
    """Returns the creation date or upload time of the file.

    Returns:
      The creation date or upload time of the file.
    """
    return self.__creation

  @property
  def size(self):
    """Returns the size of the uncompressed file.

    Returns:
      The size of the uncompressed file.
    """
    return self.__size

  @property
  def md5_hash(self):
    """Returns the MD5 hash of the uploaded file.

    Returns:
      The hash value for the uploaded file.
    """
    return self.__md5_hash

  @property
  def gs_object_name(self):
    """Returns the name of the file that was written to Cloud Storage.

    Returns:
      The name of the file that was written to Cloud Storage.
    """
    return self.__gs_object_name


def parse_file_info(field_storage):
  """Parses a `FileInfo` record from file upload `field_storage`.

  Args:
    field_storage: `cgi.FieldStorage` that represents the uploaded file.

  Returns:
    `FileInfo` record as parsed from the `field_storage` instance. This method
    will return `None` if `field_storage` was not specified.

  Raises:
    FileInfoParseError: If `field_storage` does not contain enough information
        to construct a `FileInfo` object.
  """
  info = _parse_upload_info(field_storage, FileInfoParseError)

  if info is None:
    return None

  info.pop('blob_key', None)

  return FileInfo(**info)


class BlobReferenceProperty(db.Property):
  """Property compatible with `db.Model` classes.

  Add references to blobs to domain models using BlobReferenceProperty::

      class Picture(db.Model):
        title = db.StringProperty()
        image = blobstore.BlobReferenceProperty()
        thumbnail = blobstore.BlobReferenceProperty()


  To find the size of a picture using this model::

      picture = Picture.get(picture_key)
      print picture.image.size


  `BlobInfo` objects are lazily loaded, so iterating over models for `BlobKeys`
  is efficient. The following sample code does not need to hit Datastore for
  each image key::

      list_of_untitled_blobs = []
      for picture in Picture.gql("WHERE title=''"):
        list_of_untitled_blobs.append(picture.image.key())

  """

  data_type = BlobInfo

  def get_value_for_datastore(self, model_instance):
    """Returns a model property translated to a datastore value.

    Args:
      model_instance: The model property that you want to translate.

    Returns:
      The model property that was translated from datastore.
    """
    blob_info = super(BlobReferenceProperty,
                      self).get_value_for_datastore(model_instance)
    if blob_info is None:
      return None
    return blob_info.key()

  def make_value_from_datastore(self, value):
    """Returns a datastore value to `BlobInfo`.

    Args:
      value: The datastore value that you want to translate.

    Returns:
      A `BlobInfo` that was translated from datastore.
    """
    if value is None:
      return None
    return BlobInfo(value)

  def validate(self, value):
    """Validates that an assigned value is `BlobInfo`.

    This method automatically converts from strings and `BlobKey` instances.

    Args:
      value: The value that you want to validate.

    Returns:
      Information about whether an assigned value is `BlobInfo`.
    """
    if isinstance(value, (six.string_types)):
      value = BlobInfo(BlobKey(value))
    elif isinstance(value, BlobKey):
      value = BlobInfo(value)
    return super(BlobReferenceProperty, self).validate(value)


def fetch_data(blob, start_index, end_index, rpc=None):
  """Fetches data for a blob.

  Fetches a fragment of a blob up to `MAX_BLOB_FETCH_SIZE` in length. Attempting
  to fetch a fragment that extends beyond the boundaries of the blob will return
  the amount of data from `start_index` until the end of the blob, which will be
  a smaller size than requested. Requesting a fragment that is entirely outside
  the boundaries of the blob will return an empty string. Attempting to fetch a
  negative index will raise an exception.

  Args:
    blob: A `BlobInfo`, `BlobKey`, string, or Unicode representation of
        the `BlobKey` of the blob from which you want to fetch data.
    start_index: The start index of blob data to fetch. This value must not be
        negative.
    end_index: The end index (inclusive) of the blob data to fetch. This value
        must be greater than or equal to `start_index`.
    rpc: Optional UserRPC object.

  Returns:
    A string that contains partial data of a blob. If the indexes are legal but
    outside of the boundaries of the blob, an empty string is returned.

  Raises:
    TypeError: If `start_index` or `end_index` are not indexes, or if `blob` is
        not a string, `BlobKey` or `BlobInfo`.
    DataIndexOutOfRangeError: If `start_index` is set to a value that is less
        than 0 or `end_index` is less than `start_index`.
    BlobFetchSizeTooLargeError: If the requested blob fragment is larger than
        `MAX_BLOB_FETCH_SIZE`.
    BlobNotFoundError: If the blob does not exist.
  """
  rpc = fetch_data_async(blob, start_index, end_index, rpc=rpc)
  return rpc.get_result()


def fetch_data_async(blob, start_index, end_index, rpc=None):
  """Asynchronously fetches data for a blob.

  Fetches a fragment of a blob up to `MAX_BLOB_FETCH_SIZE` in length. Attempting
  to fetch a fragment that extends beyond the boundaries of the blob will return
  the amount of data from `start_index` until the end of the blob, which will be
  a smaller size than requested. Requesting a fragment that is entirely outside
  the boundaries of the blob will return an empty string. Attempting to fetch a
  negative index will raise an exception.

  Args:
    blob: A `BlobInfo`, `BlobKey`, string, or Unicode representation of
        the `BlobKey` of the blob from which you want to fetch data.
    start_index: The start index of blob data to fetch. This value must not be
        negative.
    end_index: The end index (inclusive) of the blob data to fetch. This value
        must be greater than or equal to `start_index`.
    rpc: Optional UserRPC object.

  Returns:
    A UserRPC whose result will be a string as returned by `fetch_data()`.

  Raises:
    TypeError: If `start_index` or `end_index` are not indexes, or if `blob` is
        not a string, `BlobKey` or `BlobInfo`.
    DataIndexOutOfRangeError: If `start_index` is set to a value that is less
        than 0 or `end_index` is less than `start_index` when calling
        `rpc.get_result()`.
    BlobFetchSizeTooLargeError: If the requested blob fragment is larger than
        `MAX_BLOB_FETCH_SIZE` when calling `rpc.get_result()`.
    BlobNotFoundError: If the blob does not exist when calling
        `rpc.get_result()`.
  """
  if isinstance(blob, BlobInfo):
    blob = blob.key()
  return blobstore.fetch_data_async(blob, start_index, end_index, rpc=rpc)


class BlobReader(object):
  """Provides a read-only file-like interface to a blobstore blob."""

  SEEK_SET = 0
  SEEK_CUR = 1
  SEEK_END = 2

  def __init__(self, blob, buffer_size=131072, position=0):
    """Constructor.

    Args:
      blob: The blob key, blob info, or string blob key to read from.
      buffer_size: The minimum size to fetch chunks of data from blobstore.
      position: The initial position in the file.

    Raises:
      ValueError: If a `BlobKey`, `BlobInfo`, or string blob key is not
          supplied.
    """
    if not blob:
      raise ValueError('A BlobKey, BlobInfo or string is required.')
    if hasattr(blob, 'key'):
      self.__blob_key = blob.key()
      self.__blob_info = blob
    else:
      self.__blob_key = blob
      self.__blob_info = None
    self.__buffer_size = buffer_size
    self.__buffer = b''
    self.__position = position
    self.__buffer_position = 0
    self.__eof = False

  def __iter__(self):
    """Retrieves a file iterator for this `BlobReader`.

    Returns:
      The file iterator for the `BlobReader`.

    """
    return self

  def __getstate__(self):
    """Retrieves the serialized state for this `BlobReader`.

    Returns:
      The serialized state for the `BlobReader`.
    """
    return (self.__blob_key, self.__buffer_size, self.__position)

  def __setstate__(self, state):
    """Restores pickled state for this BlobReader."""
    self.__init__(*state)

  def close(self):
    """Closes the file.

    A closed file cannot be read or written to anymore. An operation that
    requires that the file to be open will raise a `ValueError` after the file
    has been closed. Calling `close()` more than once is allowed.
    """
    self.__blob_key = None

  def flush(self):
    raise IOError("BlobReaders are read-only")

  def next(self):
    r"""Returns the next line from the file.

    Returns:
      A string, terminated by `\\n`. The last line cannot end with `\\n`. If the
      end of the file is reached, an empty string will be returned.

    Raises:
      StopIteration: If there are no further lines to read.
    """
    line = self.readline()
    if not line:
      raise StopIteration
    return line

  def __next__(self):


    return self.next()

  def __read_from_buffer(self, size):
    """Reads at most `size` bytes from the buffer.

    Args:
      size: Number of bytes to read, or a negative value to read the entire
          buffer.

    Returns:
      Tuple (data, size):
        data: The bytes read from the buffer.
        size: The remaining unread byte count. This value is negative when
            `size` is negative. Thus when remaining size is not equal to 0, the
            calling method might choose to fill the buffer again and keep
            reading.
    """

    if not self.__blob_key:
      raise ValueError("File is closed")

    if size < 0:
      end_pos = len(self.__buffer)
    else:
      end_pos = self.__buffer_position + size
    data = self.__buffer[self.__buffer_position:end_pos]


    data_length = len(data)
    size -= data_length
    self.__position += data_length
    self.__buffer_position += data_length


    if self.__buffer_position == len(self.__buffer):
      self.__buffer = b''
      self.__buffer_position = 0

    return data, size

  def __fill_buffer(self, size=0):
    """Fills the internal buffer.

    Args:
      size: Number of bytes to read. Will be clamped to
          `[self.__buffer_size, MAX_BLOB_FETCH_SIZE]`.
    """
    read_size = min(max(size, self.__buffer_size), MAX_BLOB_FETCH_SIZE)

    self.__buffer = fetch_data(self.__blob_key, self.__position,
                               self.__position + read_size - 1)
    self.__buffer_position = 0
    self.__eof = len(self.__buffer) < read_size

  def read(self, size=-1):
    """Reads at most `size` bytes from the file.

    Fewer bytes are read if the read hits the end of the file before obtaining
    `size` bytes. If the `size` argument is negative or omitted, all data is
    read until the end of the file is reached. The bytes are returned as a
    string object. An empty string is returned immediately when the end of the
    file is reached.

    Calling `read()` without specifying a `size` is likely to be dangerous, as
    it might read excessive amounts of data.

    Args:
      size: Optional. The maximum number of bytes to read. When omitted,
          `read()` returns all remaining data in the file.

    Returns:
      The read data, as a string.
    """
    data_list = []
    while True:
      data, size = self.__read_from_buffer(size)
      data_list.append(data)
      if size == 0 or self.__eof:
        return b''.join(data_list)
      self.__fill_buffer(size)

  def readline(self, size=-1):
    """Reads one entire line from the file.

    A trailing newline character is kept in the string but can be absent when a
    file ends with an incomplete line. If the `size` argument is present and
    non-negative, it represents a maximum byte count, including the trailing
    newline and an incomplete line might be returned. An empty string is
    returned immediately only when the end of the file is reached.

    Args:
      size: Optional. The maximum number of bytes to read.

    Returns:
      The read data, as a string.
    """
    data_list = []
    while True:
      if size < 0:
        end_pos = len(self.__buffer)
      else:
        end_pos = self.__buffer_position + size
      newline_pos = self.__buffer.find(b'\n', self.__buffer_position, end_pos)
      if newline_pos != -1:

        data_list.append(
            self.__read_from_buffer(newline_pos
                                    - self.__buffer_position + 1)[0])
        break
      else:

        data, size = self.__read_from_buffer(size)
        data_list.append(data)
        if size == 0 or self.__eof:
          break
        self.__fill_buffer()
    return b''.join(data_list)

  def readlines(self, sizehint=None):
    """Reads until the end of the file using `readline()`.

    A list of lines read is returned.

    If the optional `sizehint` argument is present, instead of reading up to the
    end of the file, whole lines totalling approximately `sizehint` bytes are
    read, possibly after rounding up to an internal buffer size.

    Args:
      sizehint: A hint as to the maximum number of bytes to read.

    Returns:
      A list of strings, each being a single line from the file.
    """
    lines = []
    while sizehint is None or sizehint > 0:
      line = self.readline()
      if sizehint:
        sizehint -= len(line)
      if not line:

        break
      lines.append(line)
    return lines

  def seek(self, offset, whence=SEEK_SET):
    """Sets the file's current position, like stdio's `fseek()`_.

    Args:
      offset: The relative offset to seek to.
      whence: Optional; defines to what value the offset is relative. This
          argument defaults to `os.SEEK_SET` or `0` to use absolute file
          positioning; other valid values are `os.SEEK_CUR` or `1` to seek
          relative to the current position and `os.SEEK_END` or `2` to seek
          relative to the end of the file.

    .. _fseek():
       http://www.cplusplus.com/reference/cstdio/fseek/
    """
    if whence == BlobReader.SEEK_CUR:
      offset = self.__position + offset
    elif whence == BlobReader.SEEK_END:
      offset = self.blob_info.size + offset
    self.__buffer = b''
    self.__buffer_position = 0
    self.__position = offset
    self.__eof = False

  def tell(self):
    """Retrieves the file's current position, like stdio's `ftell()`_.

    Returns:
      The file's current position.

    .. _ftell():
       http://www.cplusplus.com/reference/cstdio/ftell/
    """
    return self.__position

  def truncate(self, size):
    """Raises an error if you attempt to truncate the file.

    Args:
      size: The size that you are attempting to truncate to.

    Raises:
      IOError: If you attempt to truncate a file in `BlobReader`.
    """
    raise IOError("BlobReaders are read-only")

  def write(self, str):
    """Raises an error if you attempt to write to the file.

    Args:
      str: The string that you are attempting to write.

    Raises:
      IOError: If you attempt to write to a file in `BlobReader`.
    """
    raise IOError("BlobReaders are read-only")

  def writelines(self, sequence):
    """Raises an error if you attempt to write to the file.

    Args:
      sequence: The sequence of strings that you are attempting to write.

    Raises:
      IOError: If you attempt to write lines to a file in `BlobReader`.
    """
    raise IOError("BlobReaders are read-only")

  @property
  def blob_info(self):
    """Returns the `BlobInfo` for this file.

    Returns:
      A string that contains the `BlobInfo`.
    """
    if not self.__blob_info:
      self.__blob_info = BlobInfo.get(self.__blob_key)
    return self.__blob_info

  @property
  def closed(self):
    """Determines whether a file is closed.

    Returns:
      `True` if this file is closed; opened files return `False`.
    """
    return self.__blob_key is None

  def __enter__(self):
    """Attempts to open the file.

    Returns:
      The open file.
    """
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    """Closes the file."""
    self.close()


class BlobMigrationRecord(db.Model):
  """Defines a model that records the result of a blob migration."""

  new_blob_ref = BlobReferenceProperty(indexed=False, name='new_blob_key')

  @classmethod
  def kind(cls):
    """Specifies the kind of blob that you are migrating.

    Returns:
      The kind of blob that you are migrating
    """
    return blobstore.BLOB_MIGRATION_KIND

  @classmethod
  def get_by_blob_key(cls, old_blob_key):
    """Fetches the `BlobMigrationRecord` for the given blob key.

    Args:
      old_blob_key: The blob key that was used in the previous app.

    Returns:
      A instance of `blobstore.BlobMigrationRecord` or `None`.
    """
    return cls.get_by_key_name(str(old_blob_key))

  @classmethod
  def get_new_blob_key(cls, old_blob_key):
    """Looks up the new key for a blob.

    Args:
      old_blob_key: The original blob key.

    Returns:
      The `blobstore.BlobKey` of the migrated blob.
    """
    record = cls.get_by_blob_key(old_blob_key)
    if record:
      return record.new_blob_ref.key()


def _serialize_range(start, end):
  """Return a string suitable for use as a value in a Range header.

  Args:
    start: The start of the bytes range e.g. 50.
    end: The end of the bytes range e.g. 100. This value is inclusive and may
      be None if the end of the range is not specified.

  Returns:
    Returns a string (e.g. "bytes=50-100") that represents a serialized Range
    header value.
  """
  if start < 0:
    range_str = '%d' % start
  elif end is None:
    range_str = '%d-' % start
  else:
    range_str = '%d-%d' % (start, end)
  return 'bytes=%s' % range_str


def _parse_range_value(range_value):
  """Parses a single range value from a Range header.

  Parses strings of the form "0-0", "0-", "0" and "-1" into (start, end) tuples,
  respectively, (0, 0), (0, None), (0, None), (-1, None).

  Args:
    range_value: A str containing a single range of a Range header.

  Returns:
    A tuple containing (start, end) where end is None if the range only has a
    start value.

  Raises:
    ValueError: If range_value is not a valid range.
  """
  end = None
  if range_value.startswith('-'):
    start = int(range_value)
    if start == 0:
      raise ValueError('-0 is not a valid range.')
  else:
    split_range = range_value.split('-', 1)
    start = int(split_range[0])
    if len(split_range) > 1 and split_range[1].strip():
      end = int(split_range[1])
      if start > end:
        raise ValueError('start must be <= end.')
  return (start, end)


def _parse_bytes(range_header):
  """Parses a full HTTP Range header.

  Args:
    range_header: The str value of the Range header.

  Returns:
    A tuple (units, parsed_ranges) where:
      units: A str containing the units of the Range header, e.g. "bytes".
      parsed_ranges: A list of (start, end) tuples in the form that
        _parsed_range_value returns.
  """
  try:
    parsed_ranges = []
    units, ranges = range_header.split('=', 1)
    for range_value in ranges.split(','):
      range_value = range_value.strip()
      if range_value:
        parsed_ranges.append(_parse_range_value(range_value))
    if not parsed_ranges:
      return None
    return units, parsed_ranges
  except ValueError:
    return None


def _check_ranges(start, end, use_range_set, use_range, range_header):
  """Set the range header.

  Args:
    start: As passed in from send_blob.
    end: As passed in from send_blob.
    use_range_set: Use range was explcilty set during call to send_blob.
    use_range: As passed in from send blob.
    range_header: Range header as received in HTTP request.

  Returns:
    Range header appropriate for placing in BLOB_RANGE_HEADER.

  Raises:
    ValueError if parameters are incorrect.  This happens:
      - start > end.
      - start < 0 and end is also provided.
      - end < 0
      - If index provided AND using the HTTP header, they don't match.
        This is a safeguard.
  """
  if end is not None and start is None:
    raise ValueError('May not specify end value without start.')


  use_indexes = start is not None
  if use_indexes:
    if end is not None:
      if start > end:
        raise ValueError('start must be < end.')
      elif start < 0:
        raise ValueError('end cannot be set if start < 0.')
    range_indexes = _serialize_range(start, end)


  if use_range_set and use_range and use_indexes:
    if range_header != range_indexes:
      raise ValueError('May not provide non-equivalent range indexes and '
                       'range headers: (header) %s != (indexes) %s'
                       % (range_header, range_indexes))


  if use_range and range_header is not None:
    return range_header
  elif use_indexes:
    return range_indexes
  else:
    return None


class BlobstoreDownloadHandler():
  """Base class for creating handlers that may send blobs to users."""

  __use_range_unset = object()

  def send_blob(self,
                environ,
                blob_key_or_info,
                content_type=None,
                save_as=None,
                start=None,
                end=None,
                **kwargs):
    """Send a blob-response based on a blob_key.

    Returns the correct response headers for serving a blob. If BlobInfo
    is provided and no content_type specified, will set request content type
    to BlobInfo's content type.

    Args:
      environ: a WSGI dict describing the HTTP request (See PEP 333).
      blob_key_or_info: BlobKey or BlobInfo record to serve.
      content_type: Content-type header to override when known. If not set,
        this header is not included in the returned headers.
        Web frameworks might set the 'content-type' header to a default type,
        which prevents GAE from setting a guessed 'content-type'.
        This behavior can be bypassed by setting headers['content-type'] to None
        or an empty string explicitly after this function call.
      save_as: If True, and BlobInfo record is provided, use BlobInfos
        filename to save-as.  If string is provided, use string as filename.
        If None or False, do not send as attachment.
      start: Start index of content-range to send.
      end: End index of content-range to send.  End index is inclusive.
      **kwargs:
        The `use_range` keyworded argument provides content range from the
        requests' Range header.

    Raises:
      ValueError on invalid save_as parameter.

    Returns:
      headers: a `dict` containing response headers

    """
    if set(kwargs) - _SEND_BLOB_PARAMETERS:
      invalid_keywords = []
      for keyword in kwargs:
        if keyword not in _SEND_BLOB_PARAMETERS:
          invalid_keywords.append(keyword)
      if len(invalid_keywords) == 1:
        raise TypeError('send_blob got unexpected keyword argument %s.'
                        % invalid_keywords[0])
      else:
        raise TypeError('send_blob got unexpected keyword arguments: %s'
                        % sorted(invalid_keywords))


    use_range = kwargs.get('use_range', self.__use_range_unset)
    use_range_set = use_range is not self.__use_range_unset

    headers = {}
    range_header = _check_ranges(start, end, use_range_set, use_range,
                                 environ.get('HTTP_RANGE'))

    if range_header is not None:
      headers[BLOB_RANGE_HEADER] = range_header

    if isinstance(blob_key_or_info, BlobInfo):
      blob_key = blob_key_or_info.key()
      blob_info = blob_key_or_info
    elif isinstance(blob_key_or_info, str) and blob_key_or_info.startswith(
        '/gs/'):
      blob_key = create_gs_key(blob_key_or_info)
      blob_info = None
    else:
      blob_key = blob_key_or_info
      blob_info = None

    headers[BLOB_KEY_HEADER] = str(blob_key)

    if content_type:
      if isinstance(content_type, six.text_type):
        content_type = content_type.encode('utf-8')
      headers['Content-Type'] = content_type

    def send_attachment(filename):
      if isinstance(filename, six.text_type):
        filename_utf8 = filename.encode('utf-8')
        headers['Content-Disposition'] = (
            _CONTENT_DISPOSITION_FORMAT_UTF8 %
            (filename_utf8, urllib.parse.quote(filename_utf8).encode('utf-8')))
      else:
        headers['Content-Disposition'] = (
            _CONTENT_DISPOSITION_FORMAT % filename)

    if save_as:
      if isinstance(save_as, (bytes, six.text_type)):
        send_attachment(save_as)
      elif blob_info and save_as is True:
        send_attachment(blob_info.filename)
      else:
        if not blob_info:
          raise ValueError('Expected BlobInfo value for blob_key_or_info.')
        else:
          raise ValueError('Unexpected value for save_as.')

    return headers

  def get_range(self, environ):
    """Get range from header if it exists.

    A range header of "bytes: 0-100" would return (0, 100).
    Args:
      environ: a WSGI dict describing the HTTP request (See PEP 333).
    Returns:
      Tuple (start, end):
        start: Start index.  None if there is None.
        end: End index (inclusive).  None if there is None.
      None if there is no request header.

    Raises:
      UnsupportedRangeFormatError: If the range format in the header is
        valid, but not supported.
      RangeFormatError: If the range format in the header is not valid.
    """
    range_header = environ.get('HTTP_RANGE')
    if range_header is None:
      return None

    parsed_range = _parse_bytes(range_header)
    if parsed_range is None:
      raise RangeFormatError('Invalid range header: %s' % range_header)

    units, ranges = parsed_range
    if len(ranges) != 1:
      raise UnsupportedRangeFormatError(
          'Unable to support multiple range values in Range header.')

    if units != _BYTES_UNIT:
      raise UnsupportedRangeFormatError(
          'Invalid unit in range header type: %s' % range_header)

    return ranges[0]

  def get(self, environ):
    """Override this method to handle GET requests to this WSGI handler.

    This method is called internally by __call__ if an instance of this class
    is used as a WSGI callable app, and the HTTP request being handled is a GET
    request.

    Args:
      environ: a WSGI dict describing the HTTP request (See PEP 333).
    Returns:
      response: a string containing body of the response
      status: HTTP status code of enum type http.HTTPStatus
      headers: a list of 2-tuples containing Response headers
    """
    raise NotImplementedError()

  def __call__(self, environ, start_response):
    if environ['REQUEST_METHOD'] != 'GET':
      return ('', http.HTTPStatus.METHOD_NOT_ALLOWED, [('Allow', 'GET')])

    response, status, headers = self.get(environ)
    start_response(f'{status.value} {status.phrase}', headers)
    return [response.encode('utf-8')]


class BlobstoreUploadHandler():
  """Base class for creation blob upload handlers."""

  def __init__(self):
    self.__uploads = None
    self.__file_infos = None

  def get_uploads(self, environ, field_name=None):
    """Get uploads sent to this handler.

    Args:
      environ: a WSGI dict describing the HTTP request (See PEP 333).
      field_name: Only select uploads that were sent as a specific field.

    Returns:
      A list of BlobInfo records corresponding to each upload.
      Empty list if there are no blob-info records for field_name.
    """
    if self.__uploads is None:
      self.__uploads = collections.defaultdict(list)
      form = cgi.FieldStorage(fp=environ['wsgi.input'], environ=environ)
      for value in form.list:
        if isinstance(value, cgi.FieldStorage):
          key = value.name
          if 'blob-key' in value.type_options:
            self.__uploads[key].append(parse_blob_info(value))

    if field_name:
      return list(self.__uploads.get(field_name, []))
    else:
      results = []
      for uploads in six.itervalues(self.__uploads):
        results.extend(uploads)
      return results

  def get_file_infos(self, environ, field_name=None):
    """Get the file infos associated to the uploads sent to this handler.

    Args:
      environ: a WSGI dict describing the HTTP request (See PEP 333).
      field_name: Only select uploads that were sent as a specific field.
        Specify None to select all the uploads.

    Returns:
      A list of FileInfo records corresponding to each upload.
      Empty list if there are no FileInfo records for field_name.
    """
    if self.__file_infos is None:
      self.__file_infos = collections.defaultdict(list)
      form = cgi.FieldStorage(fp=environ['wsgi.input'], environ=environ)
      for value in form.list:
        if isinstance(value, cgi.FieldStorage):
          key = value.name
          if 'blob-key' in value.type_options:
            self.__file_infos[key].append(parse_file_info(value))

    if field_name:
      return list(self.__file_infos.get(field_name, []))
    else:
      results = []
      for uploads in six.itervalues(self.__file_infos):
        results.extend(uploads)
      return results

  def post(self, environ):
    """Override this method to handle POST requests to this WSGI handler.

    This method is called internally by __call__ if an instance of this class
    is used as a WSGI callable app, and the HTTP request being handled is a POST
    request.

    Args:
      environ: a WSGI dict describing the HTTP request (See PEP 333).
    Returns:
      response: a string containing body of the response
      status: HTTP status code of enum type http.HTTPStatus
      headers: a list of 2-tuples containing Response headers
    """
    raise NotImplementedError()

  def __call__(self, environ, start_response):
    if environ['REQUEST_METHOD'] != 'POST':
      return ('', http.HTTPStatus.METHOD_NOT_ALLOWED, [('Allow', 'POST')])

    response, status, headers = self.post(environ)
    start_response(f'{status.value} {status.phrase}', headers)
    return [response.encode('utf-8')]



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#



"""GQL -- the SQL-like interface to the datastore.

Defines the GQL-based query class, which is a query mechanism
for the datastore which provides an alternative model for interacting with
data stored.
"""






import calendar
import datetime
import itertools
import logging
import os
import re
import time

import six
from six.moves import range
from six.moves import zip

from google.appengine.api import datastore
from google.appengine.api import datastore_errors
from google.appengine.api import datastore_types
from google.appengine.api import users





MultiQuery = datastore.MultiQuery


LOG_LEVEL = logging.DEBUG - 1



_EPOCH = datetime.datetime.utcfromtimestamp(0)


_EMPTY_LIST_PROPERTY_NAME = '__empty_IN_list__'

def Execute(query_string, *args, **keyword_args):
  """Execute command to parse and run the query.

  Calls the query parser code to build a proto-query which is an
  unbound query. The proto-query is then bound into a real query and
  executed.

  Args:
    query_string: properly formatted GQL query string.
    args: rest of the positional arguments used to bind numeric references in
          the query.
    keyword_args: dictionary-based arguments (for named parameters).

  Returns:
    the result of running the query with *args.
  """

  app = keyword_args.pop('_app', None)
  proto_query = GQL(query_string, _app=app)



  return proto_query.Bind(args, keyword_args).Run()







class GQL(object):
  """A GQL interface to the datastore.

  GQL is a SQL-like language which supports more object-like semantics
  in a language that is familiar to SQL users. The language supported by
  GQL will change over time, but will start off with fairly simple
  semantics.

  - reserved words are case insensitive
  - names are case sensitive

  The syntax for SELECT is fairly straightforward:

  SELECT [[DISTINCT] <property> [, <property> ...] | * | __key__ ]
    [FROM <entity>]
    [WHERE <condition> [AND <condition> ...]]
    [ORDER BY <property> [ASC | DESC] [, <property> [ASC | DESC] ...]]
    [LIMIT [<offset>,]<count>]
    [OFFSET <offset>]
    [HINT (ORDER_FIRST | FILTER_FIRST | ANCESTOR_FIRST)]
    [;]

  <condition> := <property> {< | <= | > | >= | = | != | IN} <value>
  <condition> := <property> {< | <= | > | >= | = | != | IN} CAST(<value>)
  <condition> := <property> IN (<value>, ...)
  <condition> := ANCESTOR IS <entity or key>

  Currently the parser is LL(1) because of the simplicity of the grammar
  (as it is largely predictive with one token lookahead).

  The class is implemented using some basic regular expression tokenization
  to pull out reserved tokens and then the recursive descent parser will act
  as a builder for the pre-compiled query. This pre-compiled query is then
  bound to arguments before executing the query.

  Initially, three parameter passing mechanisms are supported when calling
  Execute():

  - Positional parameters
  Execute('SELECT * FROM Story WHERE Author = :1 AND Date > :2')
  - Named parameters
  Execute('SELECT * FROM Story WHERE Author = :author AND Date > :date')
  - Literals (numbers, strings, booleans, and NULL)
  Execute('SELECT * FROM Story WHERE Author = \'James\'')

  Users are also given the option of doing type conversions to other datastore
  types (e.g. db.Email, db.GeoPt). The language provides a conversion function
  which allows the caller to express conversions of both literals and
  parameters. The current conversion operators are:
  - GEOPT(float, float)
  - USER(str)
  - KEY(kind, id/name[, kind, id/name...])
  - DATETIME(year, month, day, hour, minute, second)
  - DATETIME('YYYY-MM-DD HH:MM:SS')
  - DATE(year, month, day)
  - DATE('YYYY-MM-DD')
  - TIME(hour, minute, second)
  - TIME('HH:MM:SS')

  We will properly serialize and quote all values.

  It should also be noted that there are some caveats to the queries that can
  be expressed in the syntax. The parser will attempt to make these clear as
  much as possible, but some of the caveats include:
    - There is no OR operation. In most cases, you should prefer to use IN to
      express the idea of wanting data matching one of a set of values.
    - You cannot express inequality operators on multiple different properties
    - You can only have one != operator per query (related to the previous
      rule).
    - The IN and != operators must be used carefully because they can
      dramatically raise the amount of work done by the datastore. As such,
      there is a limit on the number of elements you can use in IN statements.
      This limit is set fairly low. Currently, a max of 30 datastore queries is
      allowed in a given GQL query. != translates into 2x the number of
      datastore queries, and IN multiplies by the number of elements in the
      clause (so having two IN clauses, one with 5 elements, the other with 6
      will cause 30 queries to occur).
    - Literals can take the form of basic types or as type-cast literals. On
      the other hand, literals within lists can currently only take the form of
      simple types (strings, integers, floats).


  SELECT * will return an iterable set of entities; SELECT __key__ will return
  an iterable set of Keys.
  """




















  TOKENIZE_REGEX = re.compile(r"""
    (?:'[^'\n\r]*')+|
    <=|>=|!=|=|<|>|
    :\w+|
    ,|
    \*|
    -?\d+(?:\.\d+)?|
    \w+(?:\.\w+)*|
    (?:"[^"\s]+")+|
    \(|\)|
    \S+
    """, re.VERBOSE | re.IGNORECASE)






  RESERVED_KEYWORDS = (frozenset(('SELECT', 'DISTINCT', 'FROM', 'WHERE', 'IN',
                                  'ANCESTOR', 'IS', 'AND', 'OR', 'NOT', 'ORDER',
                                  'BY', 'ASC', 'DESC', 'GROUP', 'LIMIT',
                                  'OFFSET', 'HINT', 'ORDER_FIRST',
                                  'FILTER_FIRST', 'ANCESTOR_FIRST')),
                       frozenset())



  MAX_ALLOWABLE_QUERIES = datastore.MAX_ALLOWABLE_QUERIES




  __ANCESTOR = -1


  _kind = None
  _keys_only = False
  __projection = None
  __distinct = False
  __has_ancestor = False
  __offset = -1
  __limit = -1
  __hint = ''

  def __init__(self, query_string, _app=None, _auth_domain=None,
               namespace=None):
    """Ctor.

    Parses the input query into the class as a pre-compiled query, allowing
    for a later call to Bind() to bind arguments as defined in the
    documentation.

    Args:
      query_string: properly formatted GQL query string.
      namespace: the namespace to use for this query.

    Raises:
      datastore_errors.BadQueryError: if the query is not parsable.
    """

    self.__app = _app

    self.__namespace = namespace

    self.__auth_domain = _auth_domain


    self.__symbols = self.TOKENIZE_REGEX.findall(query_string)
    initial_error = None
    for backwards_compatibility_mode in range(len(self.RESERVED_KEYWORDS)):
      self.__InitializeParseState()
      self.__active_reserved_words = self.__GenerateReservedWords(
          backwards_compatibility_mode)
      try:
        self.__Select()
      except datastore_errors.BadQueryError as error:
        logging.log(LOG_LEVEL, initial_error)
        if not initial_error:
          initial_error = error
      else:
        break
    else:
      raise initial_error

  def __InitializeParseState(self):

    self._kind = None
    self._keys_only = False
    self.__projection = None
    self.__distinct = False
    self.__has_ancestor = False
    self.__offset = -1
    self.__limit = -1
    self.__hint = ''



    self.__filters = {}

    self.__orderings = []
    self.__next_symbol = 0

  def Bind(self, args, keyword_args, cursor=None, end_cursor=None):
    """Bind the existing query to the argument list.

    Assumes that the input args are first positional, then a dictionary.
    So, if the query contains references to :1, :2 and :name, it is assumed
    that arguments are passed as (:1, :2, dict) where dict contains a mapping
    [name] -> value.

    Args:
      args: the arguments to bind to the object's unbound references.
      keyword_args: dictionary-based arguments (for named parameters).

    Raises:
      datastore_errors.BadArgumentError: when arguments are left unbound
        (missing from the inputs arguments) or when arguments do not match the
        expected type.

    Returns:
      The bound datastore.Query object. This may take the form of a MultiQuery
      object if the GQL query will require multiple backend queries to statisfy.
    """
    num_args = len(args)
    input_args = frozenset(range(num_args))
    used_args = set()

    queries = []
    enumerated_queries = self.EnumerateQueries(used_args, args, keyword_args)
    if enumerated_queries:
      query_count = len(enumerated_queries)
    else:
      query_count = 1

    for _ in range(query_count):
      queries.append(datastore.Query(self._kind,
                                     _app=self.__app,
                                     keys_only=self._keys_only,
                                     projection=self.__projection,
                                     distinct=self.__distinct,
                                     namespace=self.__namespace,
                                     cursor=cursor,
                                     end_cursor=end_cursor))

    logging.log(LOG_LEVEL,
                'Binding with %i positional args %s and %i keywords %s',
                len(args), args, len(keyword_args), keyword_args)

    for (identifier, condition), value_list in six.iteritems(self.__filters):
      for operator, params in value_list:
        value = self.__Operate(args, keyword_args, used_args, operator, params)
        if not self.__IsMultiQuery(condition):
          for query in queries:
            self.__AddFilterToQuery(identifier, condition, value, query)



    unused_args = input_args - used_args
    if unused_args:
      unused_values = [unused_arg + 1 for unused_arg in unused_args]
      raise datastore_errors.BadArgumentError('Unused positional arguments %s' %
                                              unused_values)


    if enumerated_queries:
      logging.log(LOG_LEVEL, 'Multiple Queries Bound: %s', enumerated_queries)



      for (query, enumerated_query) in zip(queries, enumerated_queries):
        query.update(enumerated_query)


    if self.__orderings:
      for query in queries:
        query.Order(*tuple(self.__orderings))

    if query_count > 1:


      return MultiQuery(queries, self.__orderings)
    else:
      return queries[0]

  def EnumerateQueries(self, used_args, args, keyword_args):
    """Create a list of all multi-query filter combinations required.

    To satisfy multi-query requests ("IN" and "!=" filters), multiple queries
    may be required. This code will enumerate the power-set of all multi-query
    filters.

    Args:
      used_args: set of used positional parameters (output only variable used in
        reporting for unused positional args)
      args: positional arguments referenced by the proto-query in self. This
        assumes the input is a tuple (and can also be called with a varargs
        param).
      keyword_args: dict of keyword arguments referenced by the proto-query in
        self.

    Returns:
      A list of maps [(identifier, condition) -> value] of all queries needed
      to satisfy the GQL query with the given input arguments.
    """
    enumerated_queries = []


    for (identifier, condition), value_list in six.iteritems(self.__filters):
      for operator, params in value_list:
        value = self.__Operate(args, keyword_args, used_args, operator, params)
        self.__AddMultiQuery(identifier, condition, value, enumerated_queries)

    return enumerated_queries

  def __CastError(self, operator, values, error_message):
    """Query building error for type cast operations.

    Args:
      operator: the failed cast operation
      values: value list passed to the cast operator
      error_message: string to emit as part of the 'Cast Error' string.

    Raises:
      BadQueryError and passes on an error message from the caller. Will raise
      BadQueryError on all calls.
    """
    raise datastore_errors.BadQueryError(
        'Type Cast Error: unable to cast %r with operation %s (%s)' %
        (values, operator.upper(), error_message))

  def __CastNop(self, values):
    """Return values[0] if it exists -- default for most where clauses."""
    if len(values) != 1:
      self.__CastError(values, 'nop', 'requires one and only one value')
    else:
      return values[0]

  def __CastList(self, values):
    """Return the full list of values -- only useful for IN clause."""
    if values:
      return values
    else:
      return None

  def __CastKey(self, values):
    """Cast input values to Key() class using encoded string or tuple list."""
    if not len(values) % 2:
      return datastore_types.Key.from_path(_app=self.__app,
                                           namespace=self.__namespace,
                                           *values)
    elif len(values) == 1 and isinstance(values[0], six.string_types):
      return datastore_types.Key(values[0])
    else:
      self.__CastError('KEY', values,
                       'requires an even number of operands '
                       'or a single encoded string')

  def __CastGeoPt(self, values):
    """Cast input to GeoPt() class using 2 input parameters."""
    if len(values) != 2:
      self.__CastError('GEOPT', values, 'requires 2 input parameters')
    return datastore_types.GeoPt(*values)

  def __CastUser(self, values):
    """Cast to User() class using the email address in values[0]."""
    if len(values) != 1:
      self.__CastError('user', values, 'requires one and only one value')
    elif values[0] is None:

      self.__CastError('user', values, 'must be non-null')
    else:
      return users.User(email=values[0], _auth_domain=self.__auth_domain)

  def __EncodeIfNeeded(self, value):
    """Simple helper function to create an str from possibly unicode strings.
    Args:
      value: input string (should pass as an instance of str or unicode).
    """
    if isinstance(value, six.text_type):
      return value.encode('utf8')
    else:
      return value

  def __CastDate(self, values):
    """Cast DATE values (year/month/day) from input (to datetime.datetime).

    Casts DATE input values formulated as ISO string or time tuple inputs.

    Args:
      values: either a single string with ISO time representation or 3
              integer valued date tuple (year, month, day).

    Returns:
      datetime.datetime value parsed from the input values.
    """

    if len(values) == 1:


      value = self.__EncodeIfNeeded(values[0])
      if isinstance(value, six.binary_type):
        try:
          time_tuple = time.strptime(value.decode(), '%Y-%m-%d')[0:6]
        except ValueError as err:
          self.__CastError('DATE', values, err)
      else:
        self.__CastError('DATE', values, 'Single input value not a byte')
    elif len(values) == 3:

      time_tuple = (values[0], values[1], values[2], 0, 0, 0)
    else:


      self.__CastError('DATE', values,
                       'function takes 1 string or 3 integer values')

    try:
      return datetime.datetime(*time_tuple)
    except ValueError as err:
      self.__CastError('DATE', values, err)

  def __CastTime(self, values):
    """Cast TIME values (hour/min/sec) from input (to datetime.datetime).

    Casts TIME input values formulated as ISO string or time tuple inputs.

    Args:
      values: either a single string with ISO time representation or 1-4
              integer valued time tuple (hour), (hour, minute),
              (hour, minute, second), (hour, minute, second, microsec).

    Returns:
      datetime.datetime value parsed from the input values.
    """
    if len(values) == 1:




      value = self.__EncodeIfNeeded(values[0])
      if isinstance(value, six.binary_type):
        try:
          time_tuple = time.strptime(value.decode(), '%H:%M:%S')
        except ValueError as err:
          self.__CastError('TIME', values, err)
        time_tuple = (1970, 1, 1) + time_tuple[3:]
        time_tuple = time_tuple[0:6]
      elif isinstance(value, int):


        time_tuple = (1970, 1, 1, value)
      else:
        self.__CastError('TIME', values,
                         'Single input value not a byte or integer hour')
    elif len(values) <= 4:

      time_tuple = (1970, 1, 1) + tuple(values)
    else:
      self.__CastError('TIME', values,
                       'function takes 1 to 4 integers or 1 string')

    try:
      return datetime.datetime(*time_tuple)
    except ValueError as err:
      self.__CastError('TIME', values, err)

  def __CastDatetime(self, values):
    """Cast DATETIME values (string or tuple) from input (to datetime.datetime).

    Casts DATETIME input values formulated as ISO string or datetime tuple
    inputs.

    Args:
      values: either a single string with ISO representation or 3-7
              integer valued time tuple (year, month, day, ...).

    Returns:
      datetime.datetime value parsed from the input values.
    """


    if len(values) == 1:
      value = self.__EncodeIfNeeded(values[0])
      if isinstance(value, six.binary_type):
        try:
          time_tuple = time.strptime(value.decode(), '%Y-%m-%d %H:%M:%S')[0:6]
        except ValueError as err:
          self.__CastError('DATETIME', values, err)
      else:
        self.__CastError('DATETIME', values, 'Single input value not a byte')
    else:
      time_tuple = values

    try:
      return datetime.datetime(*time_tuple)
    except ValueError as err:

      self.__CastError('DATETIME', values, err)

  def __Operate(self, args, keyword_args, used_args, operator, params):
    """Create a single output value from params using the operator string given.

    Args:
      args,keyword_args: arguments passed in for binding purposes (used in
          binding positional and keyword based arguments).
      used_args: set of numeric arguments accessed in this call.
          values are ints representing used zero-based positional arguments.
          used as an output parameter with new used arguments appended to the
          list.
      operator: string representing the operator to use 'nop' just returns
          the first value from params.
      params: parameter list to operate on (positional references, named
          references, or literals).

    Returns:
      A value which can be used as part of a GQL filter description (either a
      list of datastore types -- for use with IN, or a single datastore type --
      for use with other filters).
    """
    if not params:
      return None

    param_values = []
    for param in params:
      if isinstance(param, Literal):
        value = param.Get()
      else:
        value = self.__GetParam(param, args, keyword_args, used_args=used_args)
        if isinstance(param, int):
          used_args.add(param - 1)
        logging.log(LOG_LEVEL, 'found param for bind: %s value: %s',
                    param, value)
      param_values.append(value)

    logging.log(LOG_LEVEL, '%s Operating on values: %s',
                operator, repr(param_values))

    if operator in self.__cast_operators:
      result = self.__cast_operators[operator](self, param_values)
    else:
      self.__Error('Operation %s is invalid' % operator)

    return result

  def __IsMultiQuery(self, condition):
    """Return whether or not this condition could require multiple queries."""
    return condition.lower() in ('in', '!=')

  def __GetParam(self, param, args, keyword_args, used_args=None):
    """Get the specified parameter from the input arguments.

    If param is an index or named reference, args and keyword_args are used. If
    param is a cast operator tuple, will use __Operate to return the cast value.

    Args:
      param: represents either an id for a filter reference in the filter list
          (string or number) or a tuple (cast operator, params)
      args: positional args passed in by the user (tuple of arguments, indexed
          numerically by "param")
      keyword_args: dict of keyword based arguments (strings in "param")
      used_args: Index arguments passed from __Operate to determine which index
          references have been used. Default is None.

    Returns:
      The specified param from the input list.

    Raises:
      BadArgumentError: if the referenced argument doesn't exist or no type cast
      operator is present.
    """
    num_args = len(args)
    if isinstance(param, int):

      if param <= num_args:
        return args[param - 1]
      else:
        raise datastore_errors.BadArgumentError(
            'Missing argument for bind, requires argument #%i, '
            'but only has %i args.' % (param, num_args))
    elif isinstance(param, six.string_types):
      if param in keyword_args:
        return keyword_args[param]
      else:
        raise datastore_errors.BadArgumentError(
            'Missing named arguments for bind, requires argument %s' %
            param)
    elif isinstance(param, tuple) and len(param) == 2:
      cast_op, params = param
      return self.__Operate(args, keyword_args, used_args, cast_op, params)
    else:

      assert False, 'Unknown parameter %s' % param

  def __AddMultiQuery(self, identifier, condition, value, enumerated_queries):
    """Helper function to add a multi-query to previously enumerated queries.

    Args:
      identifier: property being filtered by this condition
      condition: filter condition (e.g. !=,in)
      value: value being bound
      enumerated_queries: in/out list of already bound queries -> expanded list
        with the full enumeration required to satisfy the condition query
    Raises:
      BadArgumentError if the filter is invalid (namely non-list with IN)
    """
    if condition.lower() in ('!=', 'in') and self._keys_only:
      raise datastore_errors.BadQueryError(
        'Keys only queries do not support IN or != filters.')

    def CloneQueries(queries, n):
      """Do a full copy of the queries and append to the end of the queries.

      Does an in-place replication of the input list and sorts the result to
      put copies next to one-another.

      Args:
        queries: list of all filters to clone
        n: number of copies to make

      Returns:
        Number of iterations needed to fill the structure
      """
      if not enumerated_queries:
        for _ in range(n):
          queries.append({})
        return 1
      else:
        old_size = len(queries)
        queries.extend([None] * old_size * (n - 1))
        for origin_index in reversed(range(old_size)):
          for copy_index in range(origin_index * n, (origin_index + 1) * n):
            queries[copy_index] = queries[origin_index].copy()
        return old_size

    if condition == '!=':
      if len(enumerated_queries) * 2 > self.MAX_ALLOWABLE_QUERIES:
        raise datastore_errors.BadArgumentError(
          'Cannot satisfy query -- too many IN/!= values.')

      num_iterations = CloneQueries(enumerated_queries, 2)
      for i in range(num_iterations):
        enumerated_queries[2 * i]['%s <' % identifier] = value
        enumerated_queries[2 * i + 1]['%s >' % identifier] = value
    elif condition.lower() == 'in':
      if not isinstance(value, list):
        raise datastore_errors.BadArgumentError('List expected for "IN" filter')

      in_list_size = len(value)
      if len(enumerated_queries) * in_list_size > self.MAX_ALLOWABLE_QUERIES:
        raise datastore_errors.BadArgumentError(
          'Cannot satisfy query -- too many IN/!= values.')

      if in_list_size == 0:

        num_iterations = CloneQueries(enumerated_queries, 1)
        for clone_num in range(num_iterations):

          enumerated_queries[clone_num][_EMPTY_LIST_PROPERTY_NAME] = True
        return

      num_iterations = CloneQueries(enumerated_queries, in_list_size)
      for clone_num in range(num_iterations):
        for value_num in range(len(value)):
          list_val = value[value_num]
          query_num = in_list_size * clone_num + value_num
          filt = '%s =' % identifier
          enumerated_queries[query_num][filt] = list_val

  def __AddFilterToQuery(self, identifier, condition, value, query):
    """Add a filter condition to a query based on the inputs.

    Args:
      identifier: name of the property (or self.__ANCESTOR for ancestors)
      condition: test condition
      value: test value passed from the caller
      query: query to add the filter to
    """


    if identifier != self.__ANCESTOR:
      filter_condition = '%s %s' % (identifier, condition)
      logging.log(LOG_LEVEL, 'Setting filter on "%s" with value "%s"',
                  filter_condition, value.__class__)
      datastore._AddOrAppend(query, filter_condition, value)

    else:
      logging.log(LOG_LEVEL, 'Setting ancestor query for ancestor %s', value)
      query.Ancestor(value)

  def Run(self, *args, **keyword_args):
    """Runs this query.

    Similar to datastore.Query.Run.
    Assumes that limit == -1 or > 0

    Args:
      args: arguments used to bind to references in the compiled query object.
      keyword_args: dictionary-based arguments (for named parameters).

    Returns:
      A list of results if a query count limit was passed.
      A result iterator if no limit was given.
    """


    bind_results = self.Bind(args, keyword_args)


    offset = self.offset()


    if self.__limit == -1:
      it = bind_results.Run()

      try:
        for _ in range(offset):
          next(it)
      except StopIteration:
        pass


      return it
    else:
      res = bind_results.Get(self.__limit, offset)



      return res

  def filters(self):
    """Return the compiled list of filters."""
    return self.__filters

  def hint(self):
    """Return the datastore hint."""
    return self.__hint

  def limit(self):
    """Return numerical result count limit."""
    return self.__limit

  def offset(self):
    """Return numerical result offset."""
    if self.__offset == -1:
      return 0
    else:
      return self.__offset

  def orderings(self):
    """Return the result ordering list."""
    return self.__orderings

  def is_keys_only(self):
    """Returns True if this query returns Keys, False if it returns Entities."""
    return self._keys_only

  def projection(self):
    """Returns the tuple of properties in the projection, or None."""
    return self.__projection

  def is_distinct(self):
    """Returns True if this query is marked as distinct."""
    return self.__distinct

  def kind(self):
    return self._kind



  @property
  def _entity(self):
    logging.error('GQL._entity is deprecated. Please use GQL.kind().')
    return self._kind


  __iter__ = Run





  __result_type_regex = re.compile(r'(\*|__key__)')
  __quoted_string_regex = re.compile(r'((?:\'[^\'\n\r]*\')+)')
  __ordinal_regex = re.compile(r':(\d+)$')
  __named_regex = re.compile(r':(\w+)$')
  __identifier_regex = re.compile(r'(\w+(?:\.\w+)*)$')



  __quoted_identifier_regex = re.compile(r'((?:"[^"\s]+")+)$')
  __conditions_regex = re.compile(r'(<=|>=|!=|=|<|>|is|in)$', re.IGNORECASE)
  __number_regex = re.compile(r'(\d+)$')
  __cast_regex = re.compile(
      r'(geopt|user|key|date|time|datetime)$', re.IGNORECASE)
  __cast_operators = {
      'geopt': __CastGeoPt,
      'user': __CastUser,
      'key': __CastKey,
      'datetime': __CastDatetime,
      'date': __CastDate,
      'time': __CastTime,
      'list': __CastList,
      'nop': __CastNop,
  }

  def __Error(self, error_message):
    """Generic query error.

    Args:
      error_message: string to emit as part of the 'Parse Error' string.

    Raises:
      BadQueryError and passes on an error message from the caller. Will raise
      BadQueryError on all calls to __Error()
    """
    if self.__next_symbol >= len(self.__symbols):
      raise datastore_errors.BadQueryError(
          'Parse Error: %s at end of string' % error_message)
    else:
      raise datastore_errors.BadQueryError(
          'Parse Error: %s at symbol %s' %
          (error_message, self.__symbols[self.__next_symbol]))

  def __Accept(self, symbol_string):
    """Advance the symbol and return true iff the next symbol matches input."""
    if self.__next_symbol < len(self.__symbols):
      logging.log(LOG_LEVEL, '\t%s', self.__symbols)
      logging.log(LOG_LEVEL, '\tExpect: %s Got: %s',
                  symbol_string, self.__symbols[self.__next_symbol].upper())
      if self.__symbols[self.__next_symbol].upper() == symbol_string:
        self.__next_symbol += 1
        return True
    return False

  def __Expect(self, symbol_string):
    """Require that the next symbol matches symbol_string, or emit an error.

    Args:
      symbol_string: next symbol expected by the caller

    Raises:
      BadQueryError if the next symbol doesn't match the parameter passed in.
    """
    if not self.__Accept(symbol_string):
      self.__Error('Unexpected Symbol: %s' % symbol_string)

  def __AcceptRegex(self, regex):
    """Advance and return the symbol if the next symbol matches the regex.

    Args:
      regex: the compiled regular expression to attempt acceptance on.

    Returns:
      The first group in the expression to allow for convenient access
      to simple matches. Requires () around some objects in the regex.
      None if no match is found.
    """
    if self.__next_symbol < len(self.__symbols):
      match_symbol = self.__symbols[self.__next_symbol]
      logging.log(LOG_LEVEL, '\taccept %s on symbol %s', regex, match_symbol)
      match = regex.match(match_symbol)
      if match:
        self.__next_symbol += 1
        if match.groups():
          matched_string = match.group(1)

        logging.log(LOG_LEVEL, '\taccepted %s', matched_string)
        return matched_string

    return None

  def __AcceptTerminal(self):
    """Accept either a single semi-colon or an empty string.

    Returns:
      True

    Raises:
      BadQueryError if there are unconsumed symbols in the query.
    """

    self.__Accept(';')

    if self.__next_symbol < len(self.__symbols):
      self.__Error('Expected no additional symbols')
    return True

  def __Select(self):
    """Consume the SELECT clause and everything that follows it.

    Assumes SELECT * to start.
    Transitions to a FROM clause.

    Returns:
      True if parsing completed okay.
    """
    self.__Expect('SELECT')
    if ('DISTINCT' in self.__active_reserved_words
        and self.__Accept('DISTINCT')):
      self.__distinct = True
    if not self.__Accept('*'):
      props = [self.__ExpectIdentifier()]
      while self.__Accept(','):
        props.append(self.__ExpectIdentifier())
      if props == ['__key__']:
        self._keys_only = True
      else:
        self.__projection = tuple(props)
    return self.__From()

  def __From(self):
    """Consume the FROM clause.

    Assumes a single well formed entity in the clause.
    Assumes FROM <Entity Name>
    Transitions to a WHERE clause.

    Returns:
      True if parsing completed okay.
    """
    if self.__Accept('FROM'):
      self._kind = self.__ExpectIdentifier()
    return self.__Where()

  def __Where(self):
    """Consume the WHERE cluase.

    These can have some recursion because of the AND symbol.

    Returns:
      True if parsing the WHERE clause completed correctly, as well as all
      subsequent clauses
    """
    if self.__Accept('WHERE'):
      return self.__FilterList()
    return self.__OrderBy()

  def __FilterList(self):
    """Consume the filter list (remainder of the WHERE clause)."""
    identifier = self.__Identifier()
    if not identifier:
      self.__Error('Invalid WHERE Identifier')

    condition = self.__AcceptRegex(self.__conditions_regex)
    if not condition:
      self.__Error('Invalid WHERE Condition')
    self.__CheckFilterSyntax(identifier, condition)

    if not self.__AddSimpleFilter(identifier, condition, self.__Reference()):

      if not self.__AddSimpleFilter(identifier, condition, self.__Literal()):

        type_cast = self.__TypeCast()
        if (not type_cast or
            not self.__AddProcessedParameterFilter(identifier, condition,
                                                   *type_cast)):
          self.__Error('Invalid WHERE Condition')


    if self.__Accept('AND'):
      return self.__FilterList()

    return self.__OrderBy()

  def __GetValueList(self, values_intended_for_list=False):
    """Read in a list of parameters from the tokens and return the list.

    Reads in a set of tokens by consuming symbols. If the returned list of
    values is not intended to be used within a list, only accepts literals,
    positional parameters, or named parameters. If the returned list of
    values is intended to be used within a list, also accepts values
    (cast operator, params) which represent a custom GAE Type and are
    returned by __TypeCast.

    Args:
      values_intended_for_list: Boolean to determine if the returned value list
          is intended to be used as values in a list or as values in a custom
          GAE Type. Defaults to False.

    Returns:
      A list of values parsed from the input, with values taking the form of
      strings (unbound, named reference), integers (unbound, positional
      reference), Literal() (bound value usable directly as part of a filter
      with no additional information), or a tuple (cast operator, params)
      representing a custom GAE Type. Or empty list if nothing was parsed.
    """
    params = []

    while True:
      reference = self.__Reference()
      if reference:
        params.append(reference)
      else:
        literal = self.__Literal()
        if literal:
          params.append(literal)
        elif values_intended_for_list:
          type_cast = self.__TypeCast(can_cast_list=False)
          if type_cast is None:
            self.__Error('Filter list value could not be cast '
                         'to a datastore type')
          params.append(type_cast)
        else:
          self.__Error('Parameter list requires literal or reference parameter')

      if not self.__Accept(','):
        break

    return params

  def __CheckFilterSyntax(self, identifier, condition):
    """Check that filter conditions are valid and throw errors if not.

    Args:
      identifier: identifier being used in comparison
      condition: string form of the comparison operator used in the filter
    """
    if identifier.lower() == 'ancestor':
      if condition.lower() == 'is':



        if self.__has_ancestor:
          self.__Error('Only one ANCESTOR IS" clause allowed')
      else:
        self.__Error('"IS" expected to follow "ANCESTOR"')
    elif condition.lower() == 'is':





      self.__Error('"IS" can only be used when comparing against "ANCESTOR"')

  def __AddProcessedParameterFilter(self, identifier, condition,
                                    operator, parameters):
    """Add a filter with post-processing required.

    Args:
      identifier: property being compared.
      condition: comparison operation being used with the property (e.g. !=).
      operator: operation to perform on the parameters before adding the filter.
      parameters: list of bound parameters passed to 'operator' before creating
          the filter. When using the parameters as a pass-through, pass 'nop'
          into the operator field and the first value will be used unprocessed).

    Returns:
      True if the filter was okay to add.
    """
    if parameters is None:
      return False
    if parameters[0] is None:
      return False

    logging.log(LOG_LEVEL, 'Adding Filter %s %s %s',
                identifier, condition, repr(parameters))
    filter_rule = (identifier, condition)
    if identifier.lower() == 'ancestor':
      self.__has_ancestor = True
      filter_rule = (self.__ANCESTOR, 'is')
      assert condition.lower() == 'is'


    if operator == 'list' and condition.lower() != 'in':
      self.__Error('Only IN can process a list of values')

    self.__filters.setdefault(filter_rule, []).append((operator, parameters))
    return True

  def __AddSimpleFilter(self, identifier, condition, parameter):
    """Add a filter to the query being built (no post-processing on parameter).

    Args:
      identifier: identifier being used in comparison
      condition: string form of the comparison operator used in the filter
      parameter: ID of the reference being made or a value of type Literal

    Returns:
      True if the filter could be added.
      False otherwise.
    """
    return self.__AddProcessedParameterFilter(identifier, condition,
                                              'nop', [parameter])

  def __Identifier(self):
    """Consume an identifier and return it.

    Returns:
      The identifier string. If quoted, the surrounding quotes are stripped.
    """
    logging.log(LOG_LEVEL, 'Try Identifier')
    identifier = self.__AcceptRegex(self.__identifier_regex)
    if identifier:
      if identifier.upper() in self.__active_reserved_words:
        self.__next_symbol -= 1
        self.__Error('Identifier is a reserved keyword')
    else:


      identifier = self.__AcceptRegex(self.__quoted_identifier_regex)
      if identifier:




        identifier = identifier[1:-1].replace('""', '"')
    return identifier

  def __ExpectIdentifier(self):
    id = self.__Identifier()
    if not id:
      self.__Error('Identifier Expected')
    return id

  def __Reference(self):
    """Consume a parameter reference and return it.

    Consumes a reference to a positional parameter (:1) or a named parameter
    (:email). Only consumes a single reference (not lists).

    Returns:
      The name of the reference (integer for positional parameters or string
      for named parameters) to a bind-time parameter.
    """
    logging.log(LOG_LEVEL, 'Try Reference')
    reference = self.__AcceptRegex(self.__ordinal_regex)
    if reference:

      return int(reference)
    else:
      reference = self.__AcceptRegex(self.__named_regex)
      if reference:
        return reference

    return None

  def __Literal(self):
    """Parse literals from our token list.

    Returns:
      The parsed literal from the input string (currently either a string,
      integer, or floating point value).
    """
    logging.log(LOG_LEVEL, 'Try Literal')

    literal = None

    if self.__next_symbol < len(self.__symbols):
      try:
        literal = int(self.__symbols[self.__next_symbol])
      except ValueError:
        pass
      else:
        self.__next_symbol += 1




      if literal is None:
        try:
          literal = float(self.__symbols[self.__next_symbol])
        except ValueError:
          pass
        else:
          self.__next_symbol += 1

    if literal is None:





      literal = self.__AcceptRegex(self.__quoted_string_regex)
      if literal:
        literal = literal[1:-1].replace("''", "'")

    if literal is None:


      if self.__Accept('TRUE'):
        literal = True
      elif self.__Accept('FALSE'):
        literal = False

    if literal is not None:
      return Literal(literal)



    if self.__Accept('NULL'):
      return Literal(None)
    else:
      return None

  def __TypeCast(self, can_cast_list=True):
    """Check if the next operation is a type-cast and return the cast if so.

    Casting operators look like simple function calls on their parameters. This
    code returns the cast operator found and the list of parameters provided by
    the user to complete the cast operation.

    In the case of a list, we allow the call to __GetValueList to also accept
    custom GAE Types as values.

    Args:
      can_cast_list: Boolean to determine if list can be returned as one of the
          cast operators. Default value is True.

    Returns:
      A tuple (cast operator, params) which represents the cast operation
      requested and the parameters parsed from the cast clause.

      None - if there is no TypeCast function or list is not allowed to be cast.
    """



    logging.log(LOG_LEVEL, 'Try Type Cast')
    cast_op = self.__AcceptRegex(self.__cast_regex)
    if not cast_op:
      if can_cast_list and self.__Accept('('):

        cast_op = 'list'
      else:
        return None
    else:
      cast_op = cast_op.lower()
      self.__Expect('(')

    params = self.__GetValueList(values_intended_for_list=(cast_op == 'list'))
    self.__Expect(')')

    logging.log(LOG_LEVEL, 'Got casting operator %s with params %s',
                cast_op, repr(params))

    return (cast_op, params)

  def __OrderBy(self):
    """Consume the ORDER BY clause."""
    if self.__Accept('ORDER'):
      self.__Expect('BY')
      return self.__OrderList()
    return self.__Limit()

  def __OrderList(self):
    """Consume variables and sort order for ORDER BY clause."""


    identifier = self.__Identifier()
    if identifier:
      if self.__Accept('DESC'):
        self.__orderings.append((identifier, datastore.Query.DESCENDING))
      elif self.__Accept('ASC'):
        self.__orderings.append((identifier, datastore.Query.ASCENDING))
      else:
        self.__orderings.append((identifier, datastore.Query.ASCENDING))
    else:
      self.__Error('Invalid ORDER BY Property')

    logging.log(LOG_LEVEL, self.__orderings)
    if self.__Accept(','):
      return self.__OrderList()
    return self.__Limit()

  def __Limit(self):
    """Consume the LIMIT clause."""
    if self.__Accept('LIMIT'):

      maybe_limit = self.__AcceptRegex(self.__number_regex)

      if maybe_limit:

        if self.__Accept(','):
          self.__offset = int(maybe_limit)
          if self.__offset < 0:


            self.__Error('Bad offset in LIMIT Value')
          else:
            logging.log(LOG_LEVEL, 'Set offset to %i', self.__offset)
            maybe_limit = self.__AcceptRegex(self.__number_regex)


        self.__limit = int(maybe_limit)
        if self.__limit < 1:


          self.__Error('Bad Limit in LIMIT Value')
        else:
          logging.log(LOG_LEVEL, 'Set limit to %i', self.__limit)
      else:
        self.__Error('Non-number limit in LIMIT clause')

    return self.__Offset()

  def __Offset(self):
    """Consume the OFFSET clause."""
    if self.__Accept('OFFSET'):
      if self.__offset != -1:
        self.__Error('Offset already defined in LIMIT clause')


      offset = self.__AcceptRegex(self.__number_regex)

      if offset:

        self.__offset = int(offset)
        if self.__offset < 0:


          self.__Error('Bad offset in OFFSET clause')
        else:
          logging.log(LOG_LEVEL, 'Set offset to %i', self.__offset)
      else:
        self.__Error('Non-number offset in OFFSET clause')

    return self.__Hint()

  def __Hint(self):
    """Consume the HINT clause.

    Requires one of three options (mirroring the rest of the datastore):
      HINT ORDER_FIRST
      HINT ANCESTOR_FIRST
      HINT FILTER_FIRST

    Returns:
      True if the hint clause and later clauses all parsed okay
    """
    if self.__Accept('HINT'):
      if self.__Accept('ORDER_FIRST'):
        self.__hint = 'ORDER_FIRST'
      elif self.__Accept('FILTER_FIRST'):
        self.__hint = 'FILTER_FIRST'
      elif self.__Accept('ANCESTOR_FIRST'):
        self.__hint = 'ANCESTOR_FIRST'
      else:
        self.__Error('Unknown HINT')
    return self.__AcceptTerminal()

  def __GenerateReservedWords(self, level):
    return frozenset(itertools.chain(*self.RESERVED_KEYWORDS[level:]))





class Literal(object):
  """Class for representing literal values in a way unique from unbound params.

  This is a simple wrapper class around basic types and datastore types.
  """

  def __init__(self, value):
    self.__value = value

  def Get(self):
    """Return the value of the literal."""
    return self.__value

  def __repr__(self):
    return 'Literal(%s)' % repr(self.__value)



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#



"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


_sym_db = _symbol_database.Default()


from google.appengine.base import capabilities_pb2 as google_dot_appengine_dot_base_dot_capabilities__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n:google/appengine/api/capabilities/capability_service.proto\x12\x10google.appengine\x1a(google/appengine/base/capabilities.proto\"E\n\x10IsEnabledRequest\x12\x0f\n\x07package\x18\x01 \x02(\t\x12\x12\n\ncapability\x18\x02 \x03(\t\x12\x0c\n\x04\x63\x61ll\x18\x03 \x03(\t\"\x9f\x02\n\x11IsEnabledResponse\x12I\n\x0esummary_status\x18\x01 \x01(\x0e\x32\x31.google.appengine.IsEnabledResponse.SummaryStatus\x12\x1c\n\x14time_until_scheduled\x18\x02 \x01(\x03\x12\x32\n\x06\x63onfig\x18\x03 \x03(\x0b\x32\".google.appengine.CapabilityConfig\"m\n\rSummaryStatus\x12\x0b\n\x07\x44\x45\x46\x41ULT\x10\x00\x12\x0b\n\x07\x45NABLED\x10\x01\x12\x14\n\x10SCHEDULED_FUTURE\x10\x02\x12\x11\n\rSCHEDULED_NOW\x10\x03\x12\x0c\n\x08\x44ISABLED\x10\x04\x12\x0b\n\x07UNKNOWN\x10\x05\x42?\n%com.google.appengine.api.capabilitiesB\x13\x43\x61pabilityServicePb\x88\x01\x01')



_ISENABLEDREQUEST = DESCRIPTOR.message_types_by_name['IsEnabledRequest']
_ISENABLEDRESPONSE = DESCRIPTOR.message_types_by_name['IsEnabledResponse']
_ISENABLEDRESPONSE_SUMMARYSTATUS = _ISENABLEDRESPONSE.enum_types_by_name['SummaryStatus']
IsEnabledRequest = _reflection.GeneratedProtocolMessageType('IsEnabledRequest', (_message.Message,), {
  'DESCRIPTOR' : _ISENABLEDREQUEST,
  '__module__' : 'google.appengine.api.capabilities.capability_service_pb2'

  })
_sym_db.RegisterMessage(IsEnabledRequest)

IsEnabledResponse = _reflection.GeneratedProtocolMessageType('IsEnabledResponse', (_message.Message,), {
  'DESCRIPTOR' : _ISENABLEDRESPONSE,
  '__module__' : 'google.appengine.api.capabilities.capability_service_pb2'

  })
_sym_db.RegisterMessage(IsEnabledResponse)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n%com.google.appengine.api.capabilitiesB\023CapabilityServicePb\210\001\001'
  _ISENABLEDREQUEST._serialized_start=122
  _ISENABLEDREQUEST._serialized_end=191
  _ISENABLEDRESPONSE._serialized_start=194
  _ISENABLEDRESPONSE._serialized_end=481
  _ISENABLEDRESPONSE_SUMMARYSTATUS._serialized_start=372
  _ISENABLEDRESPONSE_SUMMARYSTATUS._serialized_end=481

